from __future__ import annotations

from collections.abc import Callable, Sequence
import logging

from prettyqt import core, eventfilters


logger = logging.getLogger(__name__)


class EventCatcher(eventfilters.BaseEventFilter):
    ID = "eventcatcher"

    caught = core.Signal(core.QEvent)

    def __init__(
        self,
        include: core.QEvent.Type
        | core.event.TypeStr
        | Sequence[core.QEvent.Type | core.event.TypeStr]
        | None = None,
        exclude: core.QEvent.Type
        | core.event.TypeStr
        | Sequence[core.QEvent.Type | core.event.TypeStr]
        | None = None,
        do_filter: bool | Callable[[core.QEvent], bool] = False,
        **kwargs,
    ):
        super().__init__(**kwargs)
        match include:
            case str():
                self.include = [core.event.TYPE[include]]
            case Sequence():
                self.include = [
                    core.event.TYPE[i] if isinstance(i, str) else i for i in include
                ]
            case core.QEvent.Type():
                self.include = [include]
            case None:
                self.include = []
            case _:
                raise ValueError(include)
        match exclude:
            case str():
                self.exclude = [core.event.TYPE[exclude]]

            case Sequence():
                self.exclude = [
                    core.event.TYPE[i] if isinstance(i, str) else i for i in exclude
                ]
            case core.QEvent.Type():
                self.exclude = [exclude]
            case None:
                self.exclude = []
            case _:
                raise ValueError(exclude)
        self.do_filter = do_filter

    def eventFilter(self, source: core.QObject, event: core.QEvent) -> bool:
        if (not self.include or event.type() in self.include) and (
            not self.exclude or event.type() not in self.exclude
        ):
            self.caught.emit(event)
            # logger.debug(f"{source!r}: {event.type()!r}")
            val = self.do_filter(event) if callable(self.do_filter) else self.do_filter
            if not isinstance(val, bool):
                logger.warning(f"Non-bool value returned for {source!r}: {val} ")
            return bool(val)
        return False


if __name__ == "__main__":
    from prettyqt import widgets

    app = widgets.app()
    window = widgets.MainWindow()
    eventcatcher = EventCatcher()



from __future__ import annotations

from prettyqt import widgets


class GraphicsPathItem(widgets.AbstractGraphicsShapeItemMixin, widgets.QGraphicsPathItem):
    """Path item that you can add to a QGraphicsScene."""



from __future__ import annotations

from prettyqt import core


class AbstractListModelMixin(core.AbstractItemModelMixin):
    pass


class AbstractListModel(AbstractListModelMixin, core.QAbstractListModel):
    """Abstract model that can be subclassed to create one-dimensional list models."""



from __future__ import annotations

from prettyqt import charts


class LineSeries(charts.XYSeriesMixin, charts.QLineSeries):
    """QLineSeries with some custom properties."""


if __name__ == "__main__":
    from prettyqt.qt import QtCore

    line = LineSeries()
    line += QtCore.QPointF(1, 1)



from __future__ import annotations

import re

from prettyqt import core, gui
from prettyqt.utils import datatypes, get_repr


class RegularExpressionValidator(gui.ValidatorMixin, gui.QRegularExpressionValidator):
    """Used to check a string against a regular expression."""

    ID = "regular_expression"

    def __init__(self, *args, **kwargs):
        # allow passing strings as well as re.Pattern to the ctor
        match args, kwargs:
            case (str() as pat, *rest), _:
                super().__init__(core.QRegularExpression(pat), *rest, **kwargs)
            case _, {"regular_expression": str() as reg_str, **rest}:
                pat = core.QRegularExpression(reg_str)
                super().__init__(*args, regular_expression=pat, **rest)
            case (re.Pattern() as pat,), _:
                super().__init__(core.RegularExpression(pat))
            case _, _:
                super().__init__(*args, **kwargs)

    def __repr__(self):
        return get_repr(self, self.regularExpression())

    def __getstate__(self):
        return dict(pattern=core.RegularExpression(self.regularExpression()))

    def __setstate__(self, state):
        self.setRegularExpression(state["pattern"])

    def __reduce__(self):
        return type(self), (), self.__getstate__()

    def __eq__(self, other: object):
        return (
            self.regularExpression() == other.regularExpression()
            if isinstance(other, type(self))
            else False
        )

    def set_regex(
        self,
        regex: datatypes.PatternAndStringType,
        flags=None,
    ):
        self.setRegularExpression(datatypes.to_regular_expression(regex))

    def get_regex(self) -> str:
        val = self.regularExpression()
        return val.pattern()


if __name__ == "__main__":
    from prettyqt import widgets

    app = widgets.app()
    w = widgets.LineEdit()
    val = RegularExpressionValidator("test")
    val.set_regex(r"\w\d\d")
    w.set_validator(val)
    w.show()
    app.exec()



from __future__ import annotations

import os
import pathlib

from prettyqt.qt import QtCore
from prettyqt.utils import datatypes, get_repr


class TemporaryDir(QtCore.QTemporaryDir):
    """Creates a unique directory for temporary use."""

    def __getattr__(self, attr: str):
        return getattr(self.to_path(), attr)

    def __repr__(self):
        return get_repr(self, self.path())

    def __str__(self):
        return self.path()

    def __bool__(self):
        return self.isValid()

    def __fspath__(self) -> str:
        return self.path()

    def __truediv__(self, other: datatypes.PathType) -> pathlib.Path:
        current = pathlib.Path(self.path())
        return current / os.fspath(other)
        # new = current / other
        # if new.suffix == "":
        #     return core.Dir(new)
        # else:
        #     return core.File(new)

    def to_path(self) -> pathlib.Path:
        return pathlib.Path(self.path())


if __name__ == "__main__":
    temp = TemporaryDir()
    new = temp / "test"
    print(new)



from __future__ import annotations

import functools
import operator
from typing import TYPE_CHECKING

from prettyqt import core, gui, widgets


if TYPE_CHECKING:
    from collections.abc import Callable


class NumFilterWidget(widgets.Widget):
    """Widget with a button to choose from a comparison op (>=, >, etc) + a lineedit."""

    filter_changed = core.Signal(object)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.set_layout("horizontal", margin=0)
        self.lineedit = widgets.LineEdit()
        self.lineedit.set_validator("double", empty_allowed=True)
        self.op_button = widgets.ToolButton()
        self.op_button.setStyleSheet("QToolButton::menu-indicator{width:0px;}")
        self.box.add(self.op_button)
        self.box.add(self.lineedit)
        self.menu = widgets.Menu(triggered=self._on_menu_click)
        for comp in ["=", "<=", "<", ">=", ">"]:
            self.menu.add_action(comp)
        self.op_button.setMenu(self.menu)
        self.op_button.set_popup_mode("instant")
        self.op_button.setText("=")
        self.lineedit.value_changed.connect(self._on_filter_change)

    def _on_filter_change(self):
        self.filter_changed.emit(self.get_filter_fn())

    def _on_menu_click(self, action: gui.QAction):
        self.op_button.setText(action.text())
        self._on_filter_change()

    def get_filter_fn(self) -> Callable:
        val = self.lineedit.get_value()
        if not val:
            return lambda x: True
        val = float(val)
        match self.op_button.text():
            case "=":
                return functools.partial(operator.eq, val)
            case "<=":
                return functools.partial(operator.ge, val)
            case "<":
                return functools.partial(operator.gt, val)
            case ">=":
                return functools.partial(operator.le, val)
            case ">":
                return functools.partial(operator.lt, val)
            case _:
                raise ValueError


if __name__ == "__main__":
    app = widgets.app()
    widget = NumFilterWidget()
    widget.show()
    widget.filter_changed.connect(print)
    app.exec()



"""Tests for `prettyqt` package."""

import sys

import pytest

import prettyqt
from prettyqt import pdfwidgets


@pytest.mark.skipif(
    sys.platform == "linux" and prettyqt.qt.API.startswith("pyside"),
    reason="Segmentation fault",
)
def test_pdfview(qapp):
    widget = pdfwidgets.PdfView()
    assert widget.get_page_mode() == "single"
    assert widget.get_zoom_mode() == "custom"



from __future__ import annotations

from typing import Literal

from prettyqt import charts
from prettyqt.utils import bidict


TickTypeStr = Literal["dynamic", "fixed"]

TICK_TYPES: bidict[TickTypeStr, charts.QValueAxis.TickType] = bidict(
    dynamic=charts.QValueAxis.TickType.TicksDynamic,
    fixed=charts.QValueAxis.TickType.TicksFixed,
)


class ValueAxisMixin(charts.AbstractAxisMixin):
    def set_tick_type(self, tick_type: TickTypeStr | charts.QValueAxis.TickType):
        """Set the tick type of the legend.

        Args:
            tick_type: tick type for the legend
        """
        self.setTickType(TICK_TYPES.get_enum_value(tick_type))

    def get_tick_type(self) -> TickTypeStr:
        """Return current tick type.

        Returns:
            tick_type
        """
        return TICK_TYPES.inverse[self.tickType()]


class ValueAxis(ValueAxisMixin, charts.QValueAxis):
    pass



from __future__ import annotations

from prettyqt import constants, core, gui, widgets


class WidgetDelegate(widgets.StyledItemDelegate):
    ID = "widget"

    def __init__(
        self,
        role: constants.ItemDataRole = constants.USER_ROLE,
        parent: widgets.QAbstractItemView | None = None,
    ):
        super().__init__(parent)
        self._widget_role = role
        self._editors = {}
        self._cache_editors = True

    def paint(
        self,
        painter: gui.QPainter,
        option: widgets.QStyleOptionViewItem,
        index: core.ModelIndex,
    ):
        value = self._editor_for_index(index)
        if not isinstance(value, widgets.QWidget):
            super().paint(painter, option, index)
            return
        value.setGeometry(option.rect)
        if option.state & widgets.Style.StateFlag.State_MouseOver:
            painter.fillRect(option.rect, option.palette.highlight())
        if option.state & widgets.Style.StateFlag.State_Selected:
            painter.fillRect(option.rect, option.palette.highlight())
        pixmap = value.grab()
        painter.drawPixmap(option.rect, pixmap)
        # super().paint(painter, option, index)

    def _editor_for_index(self, index: core.ModelIndex) -> widgets.QWidget:
        # using index.data() does not work with PyQt6, it casts widgets to QObjects
        model = index.model()
        editor = model.data(index, self._widget_role)
        if not self._cache_editors:
            return editor
        # two reasons for caching:  1) possibly performance. 2) keeping a reference
        # perhaps it would be nicer to just cache the currently chosen editor, not sure.
        key = str((index.row(), index.column()))
        while (index := index.parent()).isValid():
            key += str((index.row(), index.column()))
        if key in self._editors:
            return self._editors[key]
        self._editors[key] = editor
        return editor

    def createEditor(
        self,
        parent: widgets.QWidget,
        option: widgets.QStyleOptionViewItem,
        index: core.ModelIndex,
    ):
        editor = self._editor_for_index(index)
        if isinstance(editor, widgets.QWidget):
            metaobj = core.MetaObject(editor.metaObject())
            editor = metaobj.copy(editor)
            editor.setParent(parent)
            editor.setAutoFillBackground(True)
            return editor
        return super().createEditor(parent, option, index)

    def setEditorData(self, editor: widgets.QWidget, index: core.ModelIndex):
        pass

    def setModelData(
        self,
        editor: widgets.QWidget,
        model: core.QAbstractItemModel,
        index: core.ModelIndex,
    ):
        orig = self._editor_for_index(index)
        if isinstance(orig, widgets.QWidget):
            metaobj = core.MetaObject(orig.metaObject())
            if user_prop := metaobj.get_user_property():
                value = user_prop.read(editor)
                user_prop.write(orig, value)

    def sizeHint(self, option: widgets.QStyleOptionViewItem, index: core.ModelIndex):
        editor = self._editor_for_index(index)
        if isinstance(editor, widgets.QWidget):
            return editor.sizeHint()
        return super().sizeHint(option, index)

    def updateEditorGeometry(
        self,
        editor: widgets.QWidget,
        option: widgets.QStyleOptionViewItem,
        index: core.ModelIndex,
    ):
        pass


if __name__ == "__main__":
    import logging

    app = widgets.app()
    table_widget = widgets.TableView()
    pb = widgets.PushButton("test", clicked=lambda: logging.info("x"))

    class Model(core.AbstractTableModel):
        def rowCount(self, index: core.ModelIndex | None = None) -> int:
            return 5

        def columnCount(self, index: core.ModelIndex | None = None) -> int:
            return 5

        def data(
            self,
            index: core.ModelIndex,
            role: constants.ItemDataRole = constants.DISPLAY_ROLE,
        ):
            match role, index.column():
                case constants.DISPLAY_ROLE | constants.USER_ROLE, 0:
                    # return pb
                    return widgets.PushButton("test", clicked=lambda: logging.info("x"))

                case constants.DISPLAY_ROLE | constants.USER_ROLE, 1:
                    return pb

        def flags(self, index: core.ModelIndex) -> constants.ItemFlag:
            return super().flags(index) | constants.IS_EDITABLE

    table_widget.set_delegate("widget")
    model = Model()
    table_widget.set_model(model)
    table_widget.set_edit_triggers("all")
    table_widget.resizeColumnsToContents()
    table_widget.resize(500, 300)
    table_widget.show()
    with app.debug_mode():
        app.exec()



from __future__ import annotations

from typing import Literal

from prettyqt.qt import QtNetwork
from prettyqt.utils import bidict, datatypes, get_repr


RawFormStr = Literal["name_and_value_only", "full"]

RAW_FORMS: bidict[RawFormStr, QtNetwork.QNetworkCookie.RawForm] = bidict(
    name_and_value_only=QtNetwork.QNetworkCookie.RawForm.NameAndValueOnly,
    full=QtNetwork.QNetworkCookie.RawForm.Full,
)


class NetworkCookie(QtNetwork.QNetworkCookie):
    """Holds one network cookie."""

    def __repr__(self):
        return get_repr(self, self.name(), self.value())

    def to_raw_form(self, full: bool = True):
        form = RAW_FORMS["full"] if full else RAW_FORMS["name_and_value_only"]
        self.toRawForm(form)

    def set_name(self, name: datatypes.ByteArrayType):
        name = datatypes.to_bytearray(name)
        self.setName(name)

    def get_name(self) -> str:
        return self.name().data().decode()

    def set_value(self, value: datatypes.ByteArrayType):
        value = datatypes.to_bytearray(value)
        self.setValue(value)

    def get_value(self) -> str:
        return self.value().data().decode()

    def set_expiration_date(self, date: datatypes.DateTimeType | None):
        date = datatypes.to_datetime(date)
        self.setExpirationDate(date)  # type: ignore


if __name__ == "__main__":
    cookie = NetworkCookie()



from __future__ import annotations

from prettyqt import core, widgets
from prettyqt.utils import get_repr


class LocaleEdit(widgets.ComboBox):
    value_changed = core.Signal(core.Locale)

    def __init__(
        self,
        locale: core.QLocale | None = None,
        object_name: str = "locale_edit",
        **kwargs,
    ):
        super().__init__(object_name=object_name, **kwargs)
        self._current_locale = core.Locale()
        for i in core.Locale.get_all_locales():
            self.addItem(i.bcp47Name())
        if locale is not None:
            self.set_current_locale(locale)
        self.currentTextChanged.connect(self.set_current_locale)

    def __repr__(self):
        return get_repr(self, self._current_locale)

    def clear(self):
        self._current_locale = core.Locale()
        super().clear()
        for i in core.Locale.get_all_locales():
            self.addItem(i.bcp47Name())

    # def _on_value_change(self):
    #     self._value = self.get_value()
    #     self.value_changed.emit(self._value)

    def set_current_locale(self, locale: core.QLocale | str):
        self._current_locale = core.Locale(locale)
        self.set_current_text(self._current_locale.bcp47Name())

    def is_valid(self) -> bool:
        return self._current_locale.isValid()

    def get_value(self) -> core.QLocale:
        return self._current_locale

    def set_value(self, value: core.QLocale | str):
        self.set_current_locale(value)

    value = core.Property(
        core.QLocale,
        get_value,
        set_value,
        user=True,
        doc="Currently chosen locale",
    )


if __name__ == "__main__":
    app = widgets.app()
    widget = LocaleEdit(window_title="Test")
    widget.set_value(core.Locale())
    widget.value_changed.connect(print)
    widget.show()
    app.exec()



from __future__ import annotations

from prettyqt import core


class ParallelAnimationGroup(core.AnimationGroupMixin, core.QParallelAnimationGroup):
    """Parallel group of animations."""

    def set_duration(self, duration: int):
        for anim in self:
            match anim:
                # I dont think we need to support nested Animations here..
                case core.QVariantAnimation():
                    anim.setDuration(duration)
                case _:
                    msg = "set_duration only works with QVariantAnimations."
                    raise TypeError(msg)



from __future__ import annotations

from prettyqt import core, gui, iconprovider, widgets
from prettyqt.utils import colors, datatypes, get_repr


class ColorChooserButton(widgets.Widget):
    value_changed = core.Signal(gui.Color)

    def __init__(
        self,
        *args,
        color: datatypes.ColorType | None = None,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        layout = self.set_layout("horizontal", margin=0)
        self.lineedit = widgets.LineEdit()
        self.lineedit.set_regex_validator(r"^#(?:[0-9a-fA-F]{6})$")
        layout.add(self.lineedit)
        action = gui.Action(icon="mdi.format-color-fill", triggered=self.choose_color)
        self.button = widgets.ToolButton()
        self.button.setDefaultAction(action)
        layout.add(self.button)
        self._current_color: gui.Color = gui.Color("white")
        if color is not None:
            self.set_current_color(color)

    def __repr__(self):
        return get_repr(self, self._current_color)

    @core.Slot()
    def choose_color(self):
        dlg = widgets.ColorDialog()
        if self._current_color:
            dlg.setCurrentColor(self._current_color)

        if dlg.exec():
            new_color = dlg.current_color()
            self.set_current_color(new_color)
            self.value_changed.emit(new_color)

    def set_current_color(self, color: datatypes.ColorType):
        self._current_color = colors.get_color(color)
        self.lineedit.set_text(self._current_color.name().upper())
        icon = iconprovider.for_color(self._current_color)
        self.button.set_icon(icon)

    def is_valid(self) -> bool:
        return self.lineedit.is_valid()

    def get_value(self) -> gui.Color:
        return self._current_color

    def set_value(self, value: datatypes.ColorType):
        self.set_current_color(value)

    current_color = core.Property(
        gui.QColor,
        get_value,
        set_value,
        user=True,
        doc="Currently selected color",
    )


if __name__ == "__main__":
    app = widgets.app()
    btn = ColorChooserButton(color=gui.Color("green"))
    btn.show()
    app.exec()



from __future__ import annotations

import logging
import os
import pathlib
import time
from typing import ClassVar

import git

from prettyqt import constants, core, itemmodels


logger = logging.getLogger(__name__)


class GitPythonTreeModel(itemmodels.TreeModel):
    """Base Tree Model to display a file tree combined with Git information.

    ```py
    model = GitPythonTreeModel(PATH_TO_GIT_FOLDER)
    table = widgets.TreeView()
    table.set_model(model)
    table.show()
    ```
    """

    SUPPORTS = os.PathLike | str | git.Tree | git.Repo
    HEADER: ClassVar = [
        "Name",
        "Absolute path",
        "Blob id",
        "Commit id",
        "Type",
        # "Repo",
        # "Path",
        "Size",
        "Hex SHA",
        "Tree ID",
        "Mode",
        "Symlink id",
        "Mime type",
        "Link mode",
        "File mode",
        # "Executable mode",
    ]

    def __init__(self, path: os.PathLike[str] | str | git.Tree | git.Repo, **kwargs):
        match path:
            case os.PathLike() | str():
                tree = git.Repo(path).tree()
            case git.Tree():
                tree = path
            case git.Repo():
                tree = path.tree()
            case _:
                raise TypeError(path)
        super().__init__(tree, **kwargs)

    def columnCount(self, parent: core.ModelIndex | None = None):
        return len(self.HEADER)

    def headerData(
        self,
        section: int,
        orientation: constants.Orientation,
        role: constants.ItemDataRole = constants.DISPLAY_ROLE,
    ) -> str | None:
        match orientation, role, section:
            case constants.HORIZONTAL, constants.DISPLAY_ROLE, _:
                return self.HEADER[section]
        return None

    def data(  # noqa: PLR0911
        self,
        index: core.ModelIndex,
        role: constants.ItemDataRole = constants.DISPLAY_ROLE,
    ):
        if not index.isValid():
            return None
        tree = self.data_by_index(index).obj
        match role, index.column():
            case constants.DISPLAY_ROLE, 0:
                return tree.name
            case constants.DISPLAY_ROLE, 1:
                return tree.abspath
            case constants.DISPLAY_ROLE, 2:
                if isinstance(tree, git.Tree):
                    return tree.blob_id
                return None
            case constants.DISPLAY_ROLE, 3:
                if isinstance(tree, git.Tree):
                    return tree.commit_id
                return None
            case constants.DISPLAY_ROLE, 4:
                return tree.type
            # case constants.DISPLAY_ROLE, 5:
            #     return tree.repo
            # case constants.DISPLAY_ROLE, 6:
            #     return tree.path
            case constants.DISPLAY_ROLE, 5:
                return tree.size
            case constants.DISPLAY_ROLE, 6:
                return tree.hexsha
            case constants.DISPLAY_ROLE, 7:
                if isinstance(tree, git.Tree):
                    return tree.tree_id
                return None
            case constants.DISPLAY_ROLE, 8:
                return tree.mode
            case constants.DISPLAY_ROLE, 9:
                if isinstance(tree, git.Tree):
                    return tree.symlink_id
                return None
            case constants.DISPLAY_ROLE, 10:
                if isinstance(tree, git.Blob):
                    return tree.mime_type
                return None
            case constants.DISPLAY_ROLE, 11:
                if isinstance(tree, git.Blob):
                    return tree.link_mode
                return None
            case constants.DISPLAY_ROLE, 12:
                if isinstance(tree, git.Blob):
                    return tree.file_mode
                return None
            # case constants.DISPLAY_ROLE, 15:
            #     if isinstance(tree, git.Blob):
            #         return tree.executabe_mode

    @classmethod
    def supports(cls, instance) -> bool:
        return isinstance(instance, git.Tree | git.Repo)

    def _fetch_object_children(
        self, item: GitPythonTreeModel.TreeItem
    ) -> list[GitPythonTreeModel.TreeItem]:
        return [self.TreeItem(obj=i) for i in item.obj.trees + item.obj.blobs]

    def _has_children(self, item: GitPythonTreeModel.TreeItem) -> bool:
        return False if isinstance(item.obj, git.Blob) else len(item.obj.trees) > 0


class GitPythonCommitTreeModel(itemmodels.TreeModel):
    """Base Tree Model to display a file tree combined with Git information.

    ```py
    model = GitPythonCommitTreeModel(PATH_TO_GIT_FOLDER)
    table = widgets.TreeView()
    table.set_model(model)
    table.show()
    ```
    """

    SUPPORTS = os.PathLike | str | git.Repo | git.Commit
    HEADER: ClassVar = [
        "Sha",
        "Absolute path",
    ]

    def __init__(
        self,
        repo: os.PathLike[str] | str | git.Tree | git.Repo | git.Commit,
        **kwargs,
    ):
        match repo:
            case os.PathLike() | str():
                commit = git.Repo(repo).head.commit
            case git.Repo():
                commit = repo.tree().head.commit
            case git.Commit():
                commit = repo
            case _:
                raise TypeError(repo)
        super().__init__(commit, **kwargs)

    def columnCount(self, parent: core.ModelIndex | None = None):
        return len(self.HEADER)

    def headerData(
        self,
        section: int,
        orientation: constants.Orientation,
        role: constants.ItemDataRole = constants.DISPLAY_ROLE,
    ) -> str | None:
        match orientation, role, section:
            case constants.HORIZONTAL, constants.DISPLAY_ROLE, _:
                return self.HEADER[section]
        return None

    def data(
        self,
        index: core.ModelIndex,
        role: constants.ItemDataRole = constants.DISPLAY_ROLE,
    ):
        if not index.isValid():
            return None
        commit = self.data_by_index(index).obj
        match role, index.column():
            case constants.DISPLAY_ROLE, 0:
                gmtime = time.gmtime(commit.committed_date)
                return time.strftime("%a, %d %b %Y %H:%M", gmtime)
            case constants.DISPLAY_ROLE, 0:
                return commit.authored_date
            case constants.DISPLAY_ROLE, 1:
                return commit.hexsha

    @classmethod
    def supports(cls, instance) -> bool:
        return isinstance(instance, git.Commit)

    def _fetch_object_children(
        self, item: GitPythonTreeModel.TreeItem
    ) -> list[GitPythonTreeModel.TreeItem]:
        return [self.TreeItem(obj=i) for i in item.obj.parents]

    def _has_children(self, item: GitPythonTreeModel.TreeItem) -> bool:
        return len(item.obj.parents) > 0


if __name__ == "__main__":
    from prettyqt import widgets

    app = widgets.app()
    view = widgets.TreeView()
    view.setRootIsDecorated(True)
    p = pathlib.Path.cwd()
    model = GitPythonCommitTreeModel(p)
    view.set_model(model)
    view.setEditTriggers(view.EditTrigger.AllEditTriggers)
    view.set_delegate("editor")
    view.resize(1000, 1000)
    view.show()
    with app.debug_mode():
        app.exec()



from __future__ import annotations

from prettyqt import core, widgets
from prettyqt.utils import get_repr


class GraphicsRectItem(widgets.AbstractGraphicsShapeItemMixin, widgets.QGraphicsRectItem):
    """Rectangle item that you can add to a QGraphicsScene."""

    def __repr__(self):
        return get_repr(self, self.get_rect())

    def get_rect(self) -> core.RectF:
        return core.RectF(self.rect())

    def serialize_fields(self):
        return dict(rect=self.get_rect())

    def __setstate__(self, state):
        super().__setstate__(state)
        self.setRect(state["rect"])



from __future__ import annotations

from prettyqt import constants, core, widgets


class StretchButtonToolBar(widgets.ToolBar):
    """A QToolBar that dynamically resizes its tool buttons to fit available space.

    This is done by setting fixed size on the button instances.
    Note: the class does not support `QWidgetAction`, separators, etc.
    """

    def resizeEvent(self, event):
        super().resizeEvent(event)
        size = event.size()
        self._do_layout(size)

    def actionEvent(self, event):
        super().actionEvent(event)
        if event.type() in [core.Event.Type.ActionAdded, core.Event.Type.ActionRemoved]:
            self._do_layout(self.size())

    def sizeHint(self):
        hint = super().sizeHint()
        width, height = hint.width(), hint.height()
        m1 = self.contentsMargins()
        m2 = self.layout().contentsMargins()
        dx1, dy1, dw1, dh1 = m1.left(), m1.top(), m1.right(), m1.bottom()
        dx2, dy2, dw2, dh2 = m2.left(), m2.top(), m2.right(), m2.bottom()
        dx, dy = dx1 + dx2, dy1 + dy2
        dw, dh = dw1 + dw2, dh1 + dh2
        count = len(self.actions())
        spacing = self.layout().spacing()
        space_spacing = max(count - 1, 0) * spacing
        if self.orientation() == constants.HORIZONTAL:
            width = int(height * 1.618) * count + space_spacing + dw + dx
        else:
            height = int(width * 1.618) * count + space_spacing + dh + dy
        return core.QSize(width, height)

    def _do_layout(self, size):
        """Layout the buttons to fit inside size."""
        geom = self.geometry()
        geom.setSize(size)

        # Adjust for margins (both the widgets and the layouts).
        geom = geom.marginsRemoved(self.contentsMargins())
        geom = geom.marginsRemoved(self.layout().contentsMargins())

        actions = self.actions()
        widgets_it = map(self.widgetForAction, actions)

        orientation = self.orientation()
        if orientation == constants.HORIZONTAL:
            widgets = sorted(widgets_it, key=lambda w: w.pos().x())
        else:
            widgets = sorted(widgets_it, key=lambda w: w.pos().y())

        spacing = self.layout().spacing()
        uniform_layout_helper(widgets, geom, orientation, spacing=spacing)


def uniform_layout_helper(items, contents_rect, expanding, spacing):
    """Set fixed sizes on 'items' so they can fill the whole space."""
    if len(items) == 0:
        return

    spacing_space = (len(items) - 1) * spacing

    if expanding == constants.HORIZONTAL:

        def setter(w, s):
            w.setFixedWidth(max(s, 0))

        space = contents_rect.width() - spacing_space
    else:

        def setter(w, s):
            w.setFixedHeight(max(s, 0))

        space = contents_rect.height() - spacing_space

    base_size = space // len(items)
    remainder = space % len(items)

    for i, item in enumerate(items):
        item_size = base_size + (1 if i < remainder else 0)
        setter(item, item_size)


if __name__ == "__main__":
    from prettyqt import gui

    app = widgets.app()
    widget = StretchButtonToolBar()
    actions = [
        gui.Action(
            text="super duper action",
            shortcut="Ctrl+A",
            tool_tip="some Tooltip text",
            icon="mdi.folder",
            triggered=lambda: print("test"),
        ),
        gui.Action(
            text="this is an action",
            shortcut="Ctrl+B",
            tool_tip="Tooltip",
            icon="mdi.folder-outline",
            checked=True,
            checkable=True,
        ),
    ]
    widget.addActions(actions)
    widget.show()
    app.exec()



"""Tests for `prettyqt` package."""

import re

import pytest

from prettyqt import gui, qt, syntaxhighlighters


@pytest.mark.skipif(qt.API == "pyside6", reason="Segfault")
def test_jsonhighlighter():
    doc = gui.QTextDocument()
    highlighter = syntaxhighlighters.JsonHighlighter(doc)
    highlighter.highlightBlock('{"a": "b"}')


def test_pythonhighlighter():
    doc = gui.QTextDocument()
    highlighter = syntaxhighlighters.PythonHighlighter(doc)
    highlighter.highlightBlock("def test(): pass")


def test_yamlhighlighter():
    doc = gui.QTextDocument()
    highlighter = syntaxhighlighters.YamlHighlighter(doc)
    highlighter.highlightBlock("---\ntest:\n  - hallo")


def test_markdownhighlighter():
    doc = gui.QTextDocument()
    highlighter = syntaxhighlighters.MarkdownHighlighter(doc)
    highlighter.highlightBlock("### Headline")


def test_regexmatchhighlighter():
    doc = gui.QTextDocument()
    highlighter = syntaxhighlighters.RegexMatchHighlighter(doc)
    text = "a123"
    prog = re.compile("[0-9]")
    spans = [m.span() for m in prog.finditer(text)]
    highlighter.set_spans(spans)
    highlighter.highlightBlock(text)



from __future__ import annotations

from prettyqt import core
from prettyqt.qt import QtSpatialAudio
from prettyqt.utils import datatypes


class AudioListener(core.ObjectMixin, QtSpatialAudio.QAudioListener):
    """Defines the position and orientation of the person listening to a sound field."""

    def set_position(self, position: datatypes.Vector3DType):
        self.setPosition(datatypes.to_vector3d(position))



# /// script
# requires-python = ">=3.11"
# dependencies = [
#   "prettyqt",
#   "pyside6",
#   "python-dateutil",
#   "packaging",
# ]
# ///

import inspect
import pathlib
from typing import Any

from PySide6.QtWidgets import QWidget

from prettyqt import custom_widgets, widgets


def create_widget_instance(widget_class: type[QWidget]) -> QWidget | None:
    """Create an instance of the given widget class.

    Args:
        widget_class: The class of the widget to instantiate.

    Returns:
        An instance of the widget if successful, None otherwise.
    """
    try:
        return widget_class()
    except Exception as e:
        print(f"Failed to create instance of {widget_class.__name__}: {e}")
        return None


def save_widget_screenshot(widget: QWidget, save_path: pathlib.Path) -> None:
    """Save a screenshot of the given widget.

    Args:
        widget: The widget to capture.
        save_path: The directory to save the screenshot in.
    """
    widget.show()
    app.processEvents()
    pixmap = widget.grab()
    filename = f"{type(widget).__name__}.png"
    pixmap.save(str(save_path / filename), "png")


def process_widget(widget_class: type[QWidget], save_path: pathlib.Path) -> None:
    """Process a single widget: create instance, save screenshot, and clean up.

    Args:
        widget_class: The class of the widget to process.
        save_path: The directory to save the screenshot in.
    """
    widget = create_widget_instance(widget_class)
    if widget is None or not isinstance(widget, widgets.WidgetMixin):
        return

    try:
        save_widget_screenshot(widget, save_path)
    except Exception as e:
        print(f"Error processing {widget_class.__name__}: {e}")
    finally:
        if hasattr(widget, "hide"):
            widget.hide()


def main() -> None:
    """Main function to process all widgets and save screenshots."""
    module_dict: dict[str, Any] = {"custom_widgets": custom_widgets}
    save_path = pathlib.Path().absolute() / "screenshots"
    save_path.mkdir(parents=True, exist_ok=True)

    for module in module_dict.values():
        class_members = inspect.getmembers(module, inspect.isclass)
        for _, widget_class in class_members:
            process_widget(widget_class, save_path)


if __name__ == "__main__":
    app = widgets.app()
    main()



from __future__ import annotations

from typing import Literal

from prettyqt import constants, widgets
from prettyqt.utils import bidict


PositionStr = Literal["left", "right", "top", "bottom"]

POSITION: bidict[PositionStr, widgets.QStyleOptionViewItem.Position] = bidict(
    left=widgets.QStyleOptionViewItem.Position.Left,
    right=widgets.QStyleOptionViewItem.Position.Right,
    top=widgets.QStyleOptionViewItem.Position.Top,
    bottom=widgets.QStyleOptionViewItem.Position.Bottom,
)

ViewItemFeatureStr = Literal[
    "none",
    "wrap_text",
    "alternate",
    "has_check_indicator",
    "has_display",
    "has_decoration",
]

VIEW_ITEM_FEATURE: bidict[ViewItemFeatureStr, widgets.QStyleOptionViewItem] = bidict(
    none=widgets.QStyleOptionViewItem.ViewItemFeature(0),  # type: ignore
    wrap_text=widgets.QStyleOptionViewItem.ViewItemFeature.WrapText,
    alternate=widgets.QStyleOptionViewItem.ViewItemFeature.Alternate,
    has_check_indicator=widgets.QStyleOptionViewItem.ViewItemFeature.HasCheckIndicator,
    has_display=widgets.QStyleOptionViewItem.ViewItemFeature.HasDisplay,
    has_decoration=widgets.QStyleOptionViewItem.ViewItemFeature.HasDecoration,
)


ViewItemPositionStr = Literal["invalid", "beginning", "middle", "end", "only_one"]

VIEW_ITEM_POSITION: bidict[
    ViewItemPositionStr, widgets.QStyleOptionViewItem.ViewItemPosition
] = bidict(
    invalid=widgets.QStyleOptionViewItem.ViewItemPosition.Invalid,
    beginning=widgets.QStyleOptionViewItem.ViewItemPosition.Beginning,
    middle=widgets.QStyleOptionViewItem.ViewItemPosition.Middle,
    end=widgets.QStyleOptionViewItem.ViewItemPosition.End,
    only_one=widgets.QStyleOptionViewItem.ViewItemPosition.OnlyOne,
)


class StyleOptionViewItem(widgets.StyleOptionMixin, widgets.QStyleOptionViewItem):
    def get_view_item_position(self) -> ViewItemPositionStr:
        return VIEW_ITEM_POSITION.inverse[self.viewItemPosition]

    def get_features(self) -> ViewItemFeatureStr:
        return VIEW_ITEM_FEATURE.get_list(self.features)

    def get_decoration_position(self) -> PositionStr:
        return POSITION.inverse[self.decorationPosition]

    def get_checkstate(self) -> constants.StateStr:
        return constants.STATE.inverse[self.checkState]

    def get_decoration_alignment(self) -> constants.AlignmentStr:
        return constants.ALIGNMENTS.inverse[self.decorationAlignment]


if __name__ == "__main__":
    item = StyleOptionViewItem()



from __future__ import annotations

import contextlib

from prettyqt import core
from prettyqt.qt import QtQuick


class QuickRenderControl(core.ObjectMixin, QtQuick.QQuickRenderControl):
    """Mechanism for rendering the Qt Quick scenegraph onto an offscreen render target."""

    @contextlib.contextmanager
    def edit_frame(self):
        self.beginFrame()
        yield None
        self.endFrame()


if __name__ == "__main__":
    item = QuickRenderControl()



from __future__ import annotations

from typing import Literal

from prettyqt import location, positioning
from prettyqt.utils import bidict


RelevanceHintStr = Literal["unspecified", "distance", "lexical_place_name"]

RELEVANCE_HINT: bidict[RelevanceHintStr, location.QPlaceSearchRequest.RelevanceHint] = (
    bidict(
        unspecified=location.QPlaceSearchRequest.RelevanceHint.UnspecifiedHint,
        distance=location.QPlaceSearchRequest.RelevanceHint.DistanceHint,
        lexical_place_name=location.QPlaceSearchRequest.RelevanceHint.LexicalPlaceNameHint,
    )
)


class PlaceSearchRequest(location.QPlaceSearchRequest):
    """Represents the set of parameters for a search request."""

    def get_visibility_scope(self) -> location.VisibilityStr:
        """Return the scope of the visibility.

        Returns:
            Visibility scope
        """
        return location.VISIBILITY.inverse[self.visibilityScope()]

    def get_categories(self) -> list[location.PlaceCategory]:
        return [location.PlaceCategory(i) for i in self.categories()]

    def get_search_area(self) -> positioning.GeoShape:
        area = self.searchArea()
        match area:
            case positioning.QGeoCircle():
                return positioning.GeoCircle(area)
            case positioning.QGeoPath():
                return positioning.GeoPath(area)
            case positioning.QGeoPolygon():
                return positioning.GeoPolygon(area)
            case positioning.QGeoRectangle():
                return positioning.GeoRectangle(area)
            case _:
                return positioning.GeoShape(area)

    def set_relevance_hint(
        self, hint: RelevanceHintStr | location.QPlaceSearchRequest.RelevanceHint
    ):
        """Set the relevance hint.

        Args:
            hint: Relevance hint
        """
        self.setRelevanceHint(RELEVANCE_HINT.get_enum_value(hint))

    def get_relevance_hint(self) -> RelevanceHintStr:
        """Return current relevance hint.

        Returns:
            Relevance hint
        """
        return RELEVANCE_HINT.inverse[self.relevanceHint()]


if __name__ == "__main__":
    request = PlaceSearchRequest()
    print(dir(request))



from __future__ import annotations

import os
from typing import TYPE_CHECKING

from prettyqt import qml
from prettyqt.utils import datatypes


if TYPE_CHECKING:
    from collections.abc import Iterator

    from prettyqt.qt import QtCore


class QmlApplicationEngine(qml.QmlEngineMixin, qml.QQmlApplicationEngine):
    """Convenient way to load an application from a single QML file."""

    def __iter__(self) -> Iterator[QtCore.QObject]:
        return iter(self.rootObjects())

    def load_data(
        self,
        data: datatypes.ByteArrayType,
        url: datatypes.UrlType | os.PathLike[str] | None = None,
    ):
        data = datatypes.to_bytearray(data)
        url = datatypes.to_local_url(url)
        self.loadData(data, url)

    def load_file(self, file: datatypes.UrlType | datatypes.PathType):
        if isinstance(file, os.PathLike):
            file = os.fspath(file)
        self.load(file)



from __future__ import annotations

from prettyqt import constants
from prettyqt.qt import QtCore


class DeadlineTimer(QtCore.QDeadlineTimer):
    """Marks a deadline in the future."""

    def set_type(self, typ: constants.TimerTypeStr | constants.TimerType):
        """Set the timer type.

        Args:
            typ: timer type
        """
        self.setTimerType(constants.TIMER_TYPE.get_enum_value(typ))

    def get_type(self) -> constants.TimerTypeStr:
        """Return current timer type.

        Returns:
            timer type
        """
        return constants.TIMER_TYPE.inverse[self.timerType()]



from __future__ import annotations

from collections.abc import MutableMapping
from typing import Literal

from prettyqt.qt import QtWebEngineCore
from prettyqt.utils import bidict, get_repr


FONT_FAMILY = bidict(
    standard=QtWebEngineCore.QWebEngineSettings.FontFamily.StandardFont,
    fixed=QtWebEngineCore.QWebEngineSettings.FontFamily.FixedFont,
    serif=QtWebEngineCore.QWebEngineSettings.FontFamily.SerifFont,
    sans_serif=QtWebEngineCore.QWebEngineSettings.FontFamily.SansSerifFont,
    cursive=QtWebEngineCore.QWebEngineSettings.FontFamily.CursiveFont,
    fantasy=QtWebEngineCore.QWebEngineSettings.FontFamily.FantasyFont,
    pictograph=QtWebEngineCore.QWebEngineSettings.FontFamily.PictographFont,
)

FontFamilyStr = Literal[
    "standard", "fixed", "serif", "sans_serif", "cursive", "fantasy", "pictograph"
]

FONT_SIZE = bidict(
    minimum=QtWebEngineCore.QWebEngineSettings.FontSize.MinimumFontSize,
    minimum_logical=QtWebEngineCore.QWebEngineSettings.FontSize.MinimumLogicalFontSize,
    default=QtWebEngineCore.QWebEngineSettings.FontSize.DefaultFontSize,
    default_fixed=QtWebEngineCore.QWebEngineSettings.FontSize.DefaultFixedFontSize,
)

FontSizeStr = Literal["minimum", "minimum_logical", "default", "default_fixed"]

mod = QtWebEngineCore.QWebEngineSettings

pol = mod.UnknownUrlSchemePolicy

UNKNOWN_URL_SCHEME_POLICY = bidict(
    disallow=pol.DisallowUnknownUrlSchemes,
    allow_from_user_interaction=pol.AllowUnknownUrlSchemesFromUserInteraction,
    allow_all=pol.AllowAllUnknownUrlSchemes,
)

UnknownUrlSchemePolicyStr = Literal[
    "disallow", "allow_from_user_interaction", "allow_all"
]

attr = mod.WebAttribute

WEB_ATTRIBUTES = bidict(
    auto_load_images=attr.AutoLoadImages,
    javascript_enabled=attr.JavascriptEnabled,
    javascript_can_open_windows=attr.JavascriptCanOpenWindows,
    javascript_can_access_clipboard=attr.JavascriptCanAccessClipboard,
    links_included_in_focus_chain=attr.LinksIncludedInFocusChain,
    local_storage_enabled=attr.LocalStorageEnabled,
    local_content_can_access_remote_urls=attr.LocalContentCanAccessRemoteUrls,
    spatial_navigation_enabled=attr.SpatialNavigationEnabled,
    local_content_can_access_file_urls=attr.LocalContentCanAccessFileUrls,
    hyperlink_auditing_enabled=attr.HyperlinkAuditingEnabled,
    scroll_animator_enabled=attr.ScrollAnimatorEnabled,
    error_page_enabled=attr.ErrorPageEnabled,
    plugins_enabled=attr.PluginsEnabled,
    full_screen_support_enabled=attr.FullScreenSupportEnabled,
    screen_capture_enabled=attr.ScreenCaptureEnabled,
    web_gl_enabled=attr.WebGLEnabled,
    accelerated_2d_canvas_enabled=attr.Accelerated2dCanvasEnabled,
    auto_load_icons_for_page=attr.AutoLoadIconsForPage,
    touch_icons_enabled=attr.TouchIconsEnabled,
    focus_on_navigation_enabled=attr.FocusOnNavigationEnabled,
    print_element_backgrounds=attr.PrintElementBackgrounds,
    allow_running_insecure_content=attr.AllowRunningInsecureContent,
    allow_geolocation_on_insecure_origins=attr.AllowGeolocationOnInsecureOrigins,
    allow_window_activation_from_javascript=attr.AllowWindowActivationFromJavaScript,
    show_scrollbars=attr.ShowScrollBars,
    playback_requires_user_gesture=attr.PlaybackRequiresUserGesture,
    javascript_can_paste=attr.JavascriptCanPaste,
    web_rtc_public_interfaces_only=attr.WebRTCPublicInterfacesOnly,
    dns_prefetch_enabled=attr.DnsPrefetchEnabled,
    pdf_viewer_enabled=mod.WebAttribute.PdfViewerEnabled,
)


WebAttributeStr = Literal[
    "auto_load_images",
    "javascript_enabled",
    "javascript_can_open_windows",
    "javascript_can_access_clipboard",
    "links_included_in_focus_chain",
    "local_storage_enabled",
    "local_content_can_access_remote_urls",
    "spatial_navigation_enabled",
    "local_content_can_access_file_urls",
    "hyperlink_auditing_enabled",
    "scroll_animator_enabled",
    "error_page_enabled",
    "plugins_enabled",
    "full_screen_support_enabled",
    "screen_capture_enabled",
    "web_gl_enabled",
    "accelerated_2d_canvas_enabled",
    "auto_load_icons_for_page",
    "touch_icons_enabled",
    "focus_on_navigation_enabled",
    "print_element_backgrounds",
    "allow_running_insecure_content",
    "allow_geolocation_on_insecure_origins",
    "allow_window_activation_from_javascript",
    "show_scrollbars",
    "playback_requires_user_gesture",
    "javascript_can_paste",
    "web_rtc_public_interfaces_only",
    "dns_prefetch_enabled",
    "pdf_viewer_enabled",
]


class WebEngineSettings(MutableMapping):
    def __init__(self, item: QtWebEngineCore.QWebEngineSettings):
        self.item = item

    def __repr__(self):
        return get_repr(self, dict(self))

    def __getattr__(self, val):
        return getattr(self.item, val)

    def __setitem__(self, index: WebAttributeStr, value: bool):
        self.item.setAttribute(WEB_ATTRIBUTES[index], value)

    def __getitem__(self, index: WebAttributeStr) -> bool:
        if index not in WEB_ATTRIBUTES:
            raise KeyError(index)
        return self.item.testAttribute(WEB_ATTRIBUTES[index])

    def __delitem__(self, index: WebAttributeStr):
        return self.item.resetAttribute(WEB_ATTRIBUTES[index])

    def __iter__(self):
        return iter(WEB_ATTRIBUTES.keys())

    def __len__(self):
        return len(WEB_ATTRIBUTES)

    def set_unknown_url_scheme_policy(self, policy: UnknownUrlSchemePolicyStr | pol):
        """Set the unknown url scheme policy.

        Args:
            policy: unknown url scheme policy
        """
        self.item.setUnknownUrlSchemePolicy(
            UNKNOWN_URL_SCHEME_POLICY.get_enum_value(policy)
        )

    def get_unknown_url_scheme_policy(self) -> UnknownUrlSchemePolicyStr:
        """Return current unknown url scheme policy.

        Returns:
            Unknown url scheme policy
        """
        return UNKNOWN_URL_SCHEME_POLICY.inverse[self.item.unknownUrlSchemePolicy()]

    def set_font_family(
        self,
        which: FontFamilyStr | QtWebEngineCore.QWebEngineSettings.FontFamily,
        family: str,
    ):
        """Set the actual font family to family for the specified generic family, which.

        Args:
            which: family to set
            family: generic family
        """
        self.item.setFontFamily(FONT_FAMILY.get_enum_value(which), family)

    def get_font_family(self, family: FontFamilyStr) -> str:
        """Return the actual font family for the specified generic font family.

        Args:
            family: generic font family

        Returns:
            Font family
        """
        return self.item.fontFamily(FONT_FAMILY[family])

    def set_font_size(
        self, typ: FontSizeStr | QtWebEngineCore.QWebEngineSettings.FontSize, size: int
    ):
        """Set the font size for type to size in pixels.

        Args:
            typ: font size type
            size: size in pixels
        """
        self.item.setFontSize(FONT_SIZE.get_enum_value(typ), size)

    def get_font_size(self, typ: FontSizeStr) -> int:
        """Return the default font size for type in pixels.

        Args:
            typ: font size type

        Returns:
            Font size
        """
        return self.item.fontSize(FONT_SIZE[typ])


if __name__ == "__main__":
    from prettyqt import webenginecore, widgets

    app = widgets.app()
    page = webenginecore.WebEnginePage()
    settings = page.get_settings()
    print(settings)
    # app.exec()



"""All editors adhere to a get/set_value interface + value_changed signal.

Additional "core" editors:
- LineEdit
- DoubleSlider
- Slider
- FontComboBox
- ComboBox
- CheckBox
- PushButton
- CalendarWidget
- DateEdit
- DateTimeEdit
- TimeEdit
- Dial
- SpinBox
- DoubleSpinBox
- KeySequenceEdit
- LCDNumber
- ListWidget
- PlainTextEdit
- TextEdit
- ProgressBar
- TextBrowser
"""

from __future__ import annotations



from __future__ import annotations

import logging
import os
import pathlib
from typing import TYPE_CHECKING

from prettyqt import core, widgets


if TYPE_CHECKING:
    from prettyqt.utils import datatypes


logger = logging.getLogger(__name__)


class FileTree(widgets.TreeView):
    def __init__(self, *args, filters: list[str] | None = None, **kwargs):
        super().__init__(*args, **kwargs)
        self.setRootIsDecorated(True)
        self.setup_dragdrop_move()
        model = widgets.FileSystemModel()
        model.resolve_sym_links(False)
        if filters:
            model.set_name_filters(filters, hide=False)
        self.set_model(model)
        self._expanded_ids = []

    def get_expanded_state(self, root_index: core.ModelIndex | None = None) -> list[str]:
        """Get a list of all expanded paths.

        Can be used to re-expand to the same state.
        """
        root = root_index or self.rootIndex()
        model = self.model()
        self._expanded_ids = []

        def _save_expanded_on_level(index: core.ModelIndex):
            if not self.isExpanded(index):
                return
            model = self.model()
            if index.isValid():
                path = model.data(index, model.Roles.FilePathRole)
                self._expanded_ids.append(path)
            for i in range(model.rowCount(index)):
                val = model.index(i, 0, index)
                _save_expanded_on_level(val)

        for i in range(model.rowCount(root)):
            _save_expanded_on_level(model.index(i, 0, root))
        return self._expanded_ids

    def set_expanded_state(
        self, state: list[str], root_index: core.ModelIndex | None = None
    ):
        """Set all indexes which correspond to given paths to expanded."""
        root = root_index or self.rootIndex()
        model = self.model()
        self._expanded_ids = state

        def _restore_expanded_on_level(index: core.ModelIndex):
            model = self.model()
            path = model.data(index, model.Roles.FilePathRole)
            if path not in self._expanded_ids:
                return
            self.setExpanded(index, True)
            if not model.hasChildren(index):
                return
            path = pathlib.Path(path)
            for it in path.iterdir():
                child_index = model.index(str(path / it))
                _restore_expanded_on_level(child_index)

        with self.updates_off():
            for i in range(model.rowCount(root)):
                _restore_expanded_on_level(model.index(i, 0))

    def set_root_path(self, path: datatypes.PathType):
        """Set tree rootpath to given path."""
        path = os.fspath(path)
        self.model().set_root_path(path)
        index = self.model().index(path)
        self.setRootIndex(index)


class BreadCrumbsToolBar(widgets.ToolBar):
    path_clicked = core.Signal(pathlib.Path)

    def set_breadcrumbs(self, path: datatypes.PathType):
        self.clear()
        path = pathlib.Path(path)
        logger.info("Setting breadrumbs to %s", path)
        if path.parents:
            action = self.add_action(
                text=str(path.parents[-1]), triggered=self._on_button_clicked
            )
            action.setData(path.parents[-1])
        for p in reversed(path.parents[:-1]):
            action = self.add_action(text=p.name, triggered=self._on_button_clicked)
            action.setData(p)
        if path.is_dir():
            action = self.add_action(text=path.name, triggered=self._on_button_clicked)
            action.setData(path)

    def _on_button_clicked(self, button):
        action = self.sender()
        path = action.data()
        self.path_clicked.emit(path)


class FileExplorer(widgets.Widget):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._file_edit = widgets.LineEdit()
        self._file_tree = FileTree(self)
        self.tool_bar = BreadCrumbsToolBar(self)

        self._file_tree.selectionModel().currentRowChanged.connect(
            self._on_current_changed
        )
        self.tool_bar.path_clicked.connect(self._on_breadcrumbs_clicked)
        self._file_edit.value_changed.connect(self._update_root)

        with widgets.VBoxLayout(self) as layout:
            layout.addWidget(self.tool_bar)
            # layout.addWidget(self._file_edit)
            layout.addWidget(self._file_tree)
        # self._update_root(curpath)

    def _on_breadcrumbs_clicked(self, path: pathlib.Path):
        index = self._file_tree.model().index(str(path))
        self._file_tree.set_current_index(index)

    def _on_current_changed(self, new, _):
        if (model := new.model()) is not None:
            path = new.data(model.Roles.FilePathRole)
            self.tool_bar.set_breadcrumbs(path)

    def _update_root(self, path: os.PathLike[str]):
        if pathlib.Path(path).exists():
            self._file_tree.setRoot(os.fspath(path))


if __name__ == "__main__":
    app = widgets.app()
    w = FileTree()
    w.set_root_path("C:/")
    w.show()
    app.exec()
    print(w.get_expanded_state())



from __future__ import annotations

from prettyqt import constants, widgets


DockWidgetArea = constants.DockWidgetArea


class AdjustingBoxLayoutDockWidget(widgets.DockWidget):
    """DockWidget adjusting its child widget QBoxLayout direction.

    The child widget layout direction is set according to dock widget area.
    The child widget MUST use a QBoxLayout.
    """

    def __init__(self, parent: widgets.QWidget | None = None, **kwargs):
        super().__init__(parent, **kwargs)
        self._current_area = DockWidgetArea.NoDockWidgetArea
        self.dockLocationChanged.connect(self._dock_location_changed)
        self.topLevelChanged.connect(self._top_level_changed)

    def setWidget(self, widget: widgets.QWidget):
        """Set the widget of this QDockWidget."""
        super().setWidget(widget)
        self._dock_location_changed(self._current_area)

    def _dock_location_changed(self, area: DockWidgetArea):
        self._current_area = area
        if (widget := self.widget()) is not None:
            if isinstance(layout := widget.layout(), widgets.QBoxLayout):
                if area in (
                    DockWidgetArea.LeftDockWidgetArea,
                    DockWidgetArea.RightDockWidgetArea,
                ):
                    direction = widgets.BoxLayout.Direction.TopToBottom
                else:
                    direction = widgets.BoxLayout.Direction.LeftToRight
                layout.setDirection(direction)
                self.resize(widget.minimumSize())
                self.adjustSize()

    def _top_level_changed(self, top_level: bool):
        if (widget := self.widget()) is not None and top_level:
            if isinstance(layout := widget.layout(), widgets.QBoxLayout):
                layout.setDirection(widgets.BoxLayout.Direction.LeftToRight)
                self.resize(widget.minimumSize())
                self.adjustSize()

    def showEvent(self, event):
        """Make sure this dock widget is raised when it is shown.

        This is useful for tabbed dock widgets.
        """
        self.raise_()


if __name__ == "__main__":
    app = widgets.app()
    mainwindow = widgets.MainWindow()
    dockwidget = AdjustingBoxLayoutDockWidget()
    textbox1 = widgets.PlainTextEdit()
    textbox2 = widgets.PlainTextEdit()
    container = widgets.Widget()
    container.set_layout("horizontal")
    container.set_layout(None)
    container.set_layout("horizontal")
    container.box.add(textbox1)
    dockwidget.setWidget(container)
    mainwindow.add_dockwidget(dockwidget)
    mainwindow.show()
    with app.debug_mode():
        app.exec()



from __future__ import annotations

from typing import Any, Literal

from prettyqt import gui
from prettyqt.utils import bidict, colors, datatypes


CoordinateModeStr = Literal["logical", "object", "stretch_to_device", "object_bounding"]

COORDINATE_MODE: bidict[CoordinateModeStr, gui.QGradient.CoordinateMode] = bidict(
    logical=gui.QGradient.CoordinateMode.LogicalMode,
    object=gui.QGradient.CoordinateMode.ObjectMode,
    stretch_to_device=gui.QGradient.CoordinateMode.StretchToDeviceMode,
    object_bounding=gui.QGradient.CoordinateMode.ObjectBoundingMode,
)

SpreadStr = Literal["pad", "repeat", "reflect"]

SPREAD: bidict[SpreadStr, gui.QGradient.Spread] = bidict(
    pad=gui.QGradient.Spread.PadSpread,
    repeat=gui.QGradient.Spread.RepeatSpread,
    reflect=gui.QGradient.Spread.ReflectSpread,
)

TypeStr = Literal["linear", "radial", "conical", "none"]

TYPE: bidict[TypeStr, gui.QGradient.Type] = bidict(
    linear=gui.QGradient.Type.LinearGradient,
    radial=gui.QGradient.Type.RadialGradient,
    conical=gui.QGradient.Type.ConicalGradient,
    none=gui.QGradient.Type.NoGradient,
)

PresetStr = Literal[
    "warm_flame",
    "night_fade",
    "spring_warmth",
    "juicy_peach",
    "young_passion",
    "lady_lips",
    "sunny_morning",
    "rainy_ashville",
    "frozen_dreams",
    "winter_neva",
    "dusty_grass",
    "tempting_azure",
    "heavy_rain",
    "amy_crisp",
    "mean_fruit",
    "deep_blue",
    "ripe_malinka",
    "cloudy_knoxville",
    "malibu_beach",
    "new_life",
    "true_sunset",
    "morpheus_den",
    "rare_wind",
    "near_moon",
    "wild_apple",
    "saint_petersburg",
    "plum_plate",
    "everlasting_sky",
    "happy_fisher",
    "blessing",
    "sharpeye_eagle",
    "ladoga_bottom",
    "lemon_gate",
    "itmeo_branding",
    "zeus_miracle",
    "old_hat",
    "star_wine",
    "happy_acid",
    "awesome_pine",
    "new_york",
    "shy_rainbow",
    "mixed_hopes",
    "fly_high",
]

PRESET: bidict[PresetStr, gui.QGradient.Preset] = bidict(
    warm_flame=gui.QGradient.Preset.WarmFlame,
    night_fade=gui.QGradient.Preset.NightFade,
    spring_warmth=gui.QGradient.Preset.SpringWarmth,
    juicy_peach=gui.QGradient.Preset.JuicyPeach,
    young_passion=gui.QGradient.Preset.YoungPassion,
    lady_lips=gui.QGradient.Preset.LadyLips,
    sunny_morning=gui.QGradient.Preset.SunnyMorning,
    rainy_ashville=gui.QGradient.Preset.RainyAshville,
    frozen_dreams=gui.QGradient.Preset.FrozenDreams,
    winter_neva=gui.QGradient.Preset.WinterNeva,
    dusty_grass=gui.QGradient.Preset.DustyGrass,
    tempting_azure=gui.QGradient.Preset.TemptingAzure,
    heavy_rain=gui.QGradient.Preset.HeavyRain,
    amy_crisp=gui.QGradient.Preset.AmyCrisp,
    mean_fruit=gui.QGradient.Preset.MeanFruit,
    deep_blue=gui.QGradient.Preset.DeepBlue,
    ripe_malinka=gui.QGradient.Preset.RipeMalinka,
    cloudy_knoxville=gui.QGradient.Preset.CloudyKnoxville,
    malibu_beach=gui.QGradient.Preset.MalibuBeach,
    new_life=gui.QGradient.Preset.NewLife,
    true_sunset=gui.QGradient.Preset.TrueSunset,
    morpheus_den=gui.QGradient.Preset.MorpheusDen,
    rare_wind=gui.QGradient.Preset.RareWind,
    near_moon=gui.QGradient.Preset.NearMoon,
    wild_apple=gui.QGradient.Preset.WildApple,
    saint_petersburg=gui.QGradient.Preset.SaintPetersburg,
    plum_plate=gui.QGradient.Preset.PlumPlate,
    everlasting_sky=gui.QGradient.Preset.EverlastingSky,
    happy_fisher=gui.QGradient.Preset.HappyFisher,
    blessing=gui.QGradient.Preset.Blessing,
    sharpeye_eagle=gui.QGradient.Preset.SharpeyeEagle,
    ladoga_bottom=gui.QGradient.Preset.LadogaBottom,
    lemon_gate=gui.QGradient.Preset.LemonGate,
    itmeo_branding=gui.QGradient.Preset.ItmeoBranding,
    zeus_miracle=gui.QGradient.Preset.ZeusMiracle,
    old_hat=gui.QGradient.Preset.OldHat,
    star_wine=gui.QGradient.Preset.StarWine,
    happy_acid=gui.QGradient.Preset.HappyAcid,
    awesome_pine=gui.QGradient.Preset.AwesomePine,
    new_york=gui.QGradient.Preset.NewYork,
    shy_rainbow=gui.QGradient.Preset.ShyRainbow,
    mixed_hopes=gui.QGradient.Preset.MixedHopes,
    fly_high=gui.QGradient.Preset.FlyHigh,
    # ...
)


class GradientMixin:
    def __setitem__(self, key: float, value: datatypes.ColorType):
        self.set_color_at(key, value)

    def serialize(self) -> dict[str, Any]:
        return dict(
            coordinate_mode=self.get_coordinate_mode(),
            spread=self.get_spread(),
            stops=self.get_stops(),
        )

    def set_color_at(self, key: float, color: datatypes.ColorType):
        super().setColorAt(key, colors.get_color(color))

    def set_coordinate_mode(self, mode: CoordinateModeStr | gui.QGradient.CoordinateMode):
        """Set the coordinate mode.

        Args:
            mode: coordinate mode
        """
        self.setCoordinateMode(COORDINATE_MODE.get_enum_value(mode))

    def get_coordinate_mode(self) -> CoordinateModeStr:
        """Return current coordinate mode.

        Returns:
            coordinate mode
        """
        return COORDINATE_MODE.inverse[self.coordinateMode()]

    def set_spread(self, method: SpreadStr | gui.QGradient.Spread):
        """Set the spread method.

        Args:
            method: spread method
        """
        self.setSpread(SPREAD.get_enum_value(method))

    def get_spread(self) -> SpreadStr:
        """Return current spread method.

        Returns:
            spread method
        """
        return SPREAD.inverse[self.spread()]

    def get_type(self) -> TypeStr:
        """Return current gradient type.

        Returns:
            gradient type
        """
        return TYPE.inverse[self.type()]

    def get_stops(self) -> list[tuple[float, gui.Color]]:
        return [(i, gui.Color(j)) for (i, j) in self.stops()]

    @classmethod
    def for_palette(cls, palette: gui.Palette, group: gui.palette.GroupStr = "active"):
        gradient = cls()
        for i, role_name in enumerate(gui.palette.ROLE):
            color = palette.get_color(role_name, group)
            gradient.setColorAt(i / len(gui.palette.ROLE), color)
        return gradient

    def change_brightness(self, factor: float):
        # still need to streamline changing brightness across the framework...
        # color.lighter returns a lighter color when arg is > 100 (same for darker)
        factor = int(factor * 100)
        for pos, color in grad.stops():
            if factor > 0:
                self.setColorAt(pos, color.lighter(factor))
            else:
                self.setColorAt(pos, color.darker(-factor))


class Gradient(GradientMixin, gui.QGradient):
    """Used in combination with QBrush to specify gradient fills."""


if __name__ == "__main__":
    grad = Gradient.for_palette(gui.Palette())



from __future__ import annotations

from typing import TYPE_CHECKING

from prettyqt import core
from prettyqt.qt import QtNetwork
from prettyqt.utils import datatypes, get_repr


if TYPE_CHECKING:
    from collections.abc import Iterator


class NetworkCookieJar(core.ObjectMixin, QtNetwork.QNetworkCookieJar):
    """Implements a simple jar of QNetworkCookie objects."""

    def __add__(self, other: QtNetwork.QNetworkCookie):
        self.insertCookie(other)
        return self

    def __getitem__(self, url: datatypes.UrlType) -> list[QtNetwork.QNetworkCookie]:
        url = core.Url(url)
        return self.cookiesForUrl(url)

    def __repr__(self):
        return get_repr(self)

    def __iter__(self) -> Iterator[QtNetwork.QNetworkCookie]:
        return iter(self.allCookies())

    def set_cookies_from_url(
        self, cookies: list[QtNetwork.QNetworkCookie], url: datatypes.UrlType
    ) -> bool:
        return self.setCookiesFromUrl(cookies, core.QUrl(url))


if __name__ == "__main__":
    jar = NetworkCookieJar()



from __future__ import annotations

import re
from typing import Literal

from prettyqt import core, gui


ModeStr = Literal["word", "all"]

PAT = re.compile("\\s+")


class SelectedWordHighlighter(gui.SyntaxHighlighter):
    def __init__(self, parent: gui.QTextDocument | None = None):
        super().__init__(parent)
        self._selection_term = ""
        self._mode = "word"
        self._highlight_format = gui.TextCharFormat()
        self._highlight_format.setBackground(gui.Color(255, 210, 120))
        self._highlight_format.setFontWeight(gui.QFont.Weight.Bold)
        self._highlight_pattern = None
        self._widget = parent.parent()
        self._widget.selectionChanged.connect(self._filter_selection_for_single_word)

    def set_selection_term(self, term: str):
        if term == self._selection_term:
            return
        self._selection_term = term
        self._update()

    def _update(self):
        match self._mode:
            case "all":
                search_pattern = self._selection_term
            case "word":
                search_pattern = f"\\b{self._selection_term}\\b"
            case _:
                raise ValueError(self._mode)
        self._highlight_pattern = core.RegularExpression(search_pattern)
        self.rehighlight()

    def get_selection_term(self) -> str:
        return self._selection_term

    def set_selection_mode(self, mode: ModeStr):
        if mode == self._mode:
            return
        self._mode = mode
        self._update()

    def get_selection_mode(self) -> ModeStr:
        return self._mode

    def highlightBlock(self, text):
        if len(self._selection_term) > 1:
            self._apply_highlight(text)

    def _apply_highlight(self, text):
        for m in self._highlight_pattern.finditer(text):
            length = m.span()[1] - m.span()[0]
            self.setFormat(m.span()[0], length, self._highlight_format)

    def _filter_selection_for_single_word(self):
        tc = self._widget.selecter.get_text_cursor()
        current_selection = tc.selectedText()
        if not PAT.split(current_selection):  # SkipEmptyParts
            self.set_selection_term("")
            return
        tc.select_text("start_of_word", "end_of_word")
        term = current_selection if current_selection == tc.selectedText() else ""
        self.set_selection_term(term)

    selectionTerm = core.Property(  # noqa: N815
        str,
        get_selection_term,
        set_selection_term,
        doc="String to highlight in the document",
    )
    selectionMode = core.Property(  # noqa: N815
        str,
        get_selection_mode,
        set_selection_mode,
        doc="Selection mode",
    )


if __name__ == "__main__":
    from prettyqt import widgets

    app = widgets.app()
    editor = widgets.PlainTextEdit()
    highlighter = SelectedWordHighlighter(editor.document())
    editor.set_syntaxhighlighter(highlighter)
    # highlighter = SelectedWordHighlighter(editor.document())
    editor.show()
    app.exec()



from __future__ import annotations

from typing import Literal

from prettyqt.qt import QtGui
from prettyqt.utils import bidict, datatypes


mod = QtGui.QImageIOHandler

ImageOptionStr = Literal[
    "size",
    "clip_rect",
    "scaled_size",
    "scaled_clip_rect",
    "description",
    "compression_ratio",
    "gamma",
    "quality",
    "name",
    "subtype",
    "incremental_reading",
    "endianness",
    "animation",
    "background_color",
    # "image_format",
    "supported_sub_types",
    "optimized_write",
    "progressive_scan_write",
    "image_transformation",
]

IMAGE_OPTION: bidict[ImageOptionStr, mod.ImageOption] = bidict(
    size=mod.ImageOption.Size,
    clip_rect=mod.ImageOption.ClipRect,
    scaled_size=mod.ImageOption.ScaledSize,
    scaled_clip_rect=mod.ImageOption.ScaledClipRect,
    description=mod.ImageOption.Description,
    compression_ratio=mod.ImageOption.CompressionRatio,
    gamma=mod.ImageOption.Gamma,
    quality=mod.ImageOption.Quality,
    name=mod.ImageOption.Name,
    subtype=mod.ImageOption.SubType,
    incremental_reading=mod.ImageOption.IncrementalReading,
    endianness=mod.ImageOption.Endianness,
    animation=mod.ImageOption.Animation,
    background_color=mod.ImageOption.BackgroundColor,
    # image_format=mod.ImageOption.ImageFormat,
    supported_sub_types=mod.ImageOption.SupportedSubTypes,
    optimized_write=mod.ImageOption.OptimizedWrite,
    progressive_scan_write=mod.ImageOption.ProgressiveScanWrite,
    image_transformation=mod.ImageOption.ImageTransformation,
)

TransformationStr = Literal[
    "none",
    "mirror",
    "flip",
    "rotate_180",
    "roate_90",
    "mirror_and_rotate_90",
    "flip_and_rotate_90",
    "rotate_270",
]

TRANSFORMATION: bidict[TransformationStr, mod.Transformation] = bidict(
    none=mod.Transformation.TransformationNone,
    mirror=mod.Transformation.TransformationMirror,
    flip=mod.Transformation.TransformationFlip,
    rotate_180=mod.Transformation.TransformationRotate180,
    roate_90=mod.Transformation.TransformationRotate90,
    mirror_and_rotate_90=mod.Transformation.TransformationMirrorAndRotate90,
    flip_and_rotate_90=mod.Transformation.TransformationFlipAndRotate90,
    rotate_270=mod.Transformation.TransformationRotate270,
)


class ImageIOHandler(mod):
    """Defines the common image I/O interface for all image formats in Qt."""

    def __getitem__(self, key: ImageOptionStr | mod.ImageOption) -> datatypes.Variant:
        return self.get_option(key)

    def __setitem__(
        self, key: ImageOptionStr | mod.ImageOption, value: datatypes.Variant
    ):
        self.set_option(key, value)

    def get_format(self) -> str:
        return self.format().data().decode()

    def set_option(
        self, option: ImageOptionStr | mod.ImageOption, value: datatypes.Variant
    ):
        """Set option to given value.

        Args:
            option: option to use
            value: value to set
        """
        self.setOption(IMAGE_OPTION.get_enum_value(option), value)

    def get_option(self, option: ImageOptionStr | mod.ImageOption) -> datatypes.Variant:
        """Return the value assigned to option.

        Args:
            option: option to get

        Returns:
            option
        """
        return self.option(IMAGE_OPTION.get_enum_value(option))

    def supports_option(self, option: ImageOptionStr | mod.ImageOption) -> bool:
        """Return whether the image handler supports given option.

        Args:
            option: option to check

        Returns:
            option
        """
        return self.supportsOption(IMAGE_OPTION.get_enum_value(option))


if __name__ == "__main__":
    writer = ImageIOHandler()



"""Tests for `prettyqt` package."""

import datetime
import pathlib

# import pickle
import sys

# import logging
import tempfile

import pytest

# import prettyqt
from prettyqt import constants, core, gui, iconprovider, widgets
from prettyqt.utils import InvalidParamError


def test_boxlayout(qtbot):
    layout = widgets.HBoxLayout(margin=0)
    widget = widgets.RadioButton("test")
    layout += widget
    layout2 = widgets.HBoxLayout()
    layout += layout2
    assert layout[1] == layout2
    layout.set_size_constraint("maximum")
    assert layout.get_size_constraint() == "maximum"
    layout.set_alignment("left")
    layout.set_alignment("left", widget)
    with pytest.raises(InvalidParamError):
        layout.set_alignment("test")
    # assert layout.get_alignment() == "left"
    with pytest.raises(InvalidParamError):
        layout.set_size_constraint("bla")
    layout.set_margin(0)
    assert len(layout) == 2
    layout.add_stretch(1)
    layout.add_spacing(1)
    layout.set_direction("bottom_to_top")
    with pytest.raises(InvalidParamError):
        layout.set_direction("test")
    assert layout.get_direction() == "bottom_to_top"


def test_buttongroup(qtbot):
    widget = widgets.ButtonGroup()
    btn = widgets.RadioButton("test")
    widget.addButton(btn, id=2)
    assert widget[2] == btn


def test_calendarwiget(qtbot):
    widget = widgets.CalendarWidget()
    qtbot.addWidget(widget)
    assert widget.get_date() == widget.get_value()
    widget.set_range(datetime.date(2000, 1, 1), datetime.date(2020, 1, 1))
    widget.set_value(datetime.date(2000, 10, 10))
    widget.set_selection_mode(None)
    widget.set_selection_mode("single")
    assert widget.get_selection_mode() == "single"
    with pytest.raises(InvalidParamError):
        widget.set_selection_mode("test")


def test_checkbox(qtbot):
    widget = widgets.CheckBox()
    qtbot.addWidget(widget)
    widget.set_disabled()
    widget.set_enabled()
    assert not bool(widget)
    widget.set_value(True)
    assert widget.get_value() is True
    with pytest.raises(InvalidParamError):
        widget.set_checkstate("test")


def test_colordialog(qtbot):
    dlg = widgets.ColorDialog()
    qtbot.addWidget(dlg)
    assert str(dlg.current_color()) == str(gui.Color("white"))


def test_combobox(qtbot):
    box = widgets.ComboBox()
    qtbot.addWidget(box)
    box.set_disabled()
    box.set_enabled()
    box.add("test", data="data", icon="mdi.timer")
    assert len(box) == 1
    box.set_insert_policy("bottom")
    assert box.get_insert_policy() == "bottom"
    with pytest.raises(InvalidParamError):
        box.set_insert_policy("bla")
    box.set_size_adjust_policy("first_show")
    with pytest.raises(InvalidParamError):
        box.set_size_adjust_policy("bla")
    assert box.get_size_adjust_policy() == "first_show"
    box.set_icon_size(10)
    assert box.get_icon_size() == core.Size(10, 10)
    box.set_min_char_length(10)
    box.add("test2", data="data", icon="mdi.timer")
    box.set_text("test2")
    assert box.text() == "test2"
    box.add_items(dict(a="b"))
    box.add_items(["a"])
    box.add_items([("a", "x")])
    box.get_value()
    box.set_value("x")


def test_commandlinkbutton(qtbot):
    widget = widgets.CommandLinkButton("Test")
    qtbot.addWidget(widget)
    widget.set_disabled()
    widget.set_enabled()
    widget.set_icon("mdi.timer")
    with pytest.raises(InvalidParamError):
        widget.set_style_icon("bla")


def test_completer(qtbot):
    completer = widgets.Completer()
    completer.set_sort_mode(None)
    completer.set_sort_mode("unsorted")
    with pytest.raises(InvalidParamError):
        completer.set_sort_mode("test")
    assert completer.get_sort_mode() == "unsorted"
    completer.set_completion_mode("popup")
    with pytest.raises(InvalidParamError):
        completer.set_completion_mode("test")
    assert completer.get_completion_mode() == "popup"
    completer.set_filter_mode("contains")
    with pytest.raises(InvalidParamError):
        completer.set_filter_mode("test")
    assert completer.get_filter_mode() == "contains"
    strings = ["a", "b"]
    completer.set_strings(strings)
    assert completer.get_completions() == strings
    completer.set_case_sensitive(True)
    completer.set_case_sensitive(False)
    assert not completer.is_case_sensitive()


def test_dateedit(qtbot):
    widget = widgets.DateEdit()
    qtbot.addWidget(widget)
    widget.set_disabled()
    widget.set_enabled()
    dt = datetime.date(2000, 11, 11)
    widget.set_value(dt)
    assert widget.get_value() == dt


def test_datetimeedit(qtbot):
    widget = widgets.DateTimeEdit()
    qtbot.addWidget(widget)
    widget.set_disabled()
    widget.set_enabled()
    dt = datetime.datetime(2000, 11, 11)
    widget.set_value(dt)
    widget.set_format("dd.MM.yyyy")
    # assert widget.get_value() == dt
    assert widget.get_section_text("day") == "11"
    with pytest.raises(InvalidParamError):
        widget.get_section_text("test")
    widget.set_current_section("day")
    with pytest.raises(InvalidParamError):
        widget.set_current_section("test")
    assert widget.get_current_section() == "day"
    widget.get_displayed_sections()


def test_dialog(qtbot, qttester):
    dlg = widgets.Dialog()
    dlg.set_layout("horizontal")
    qtbot.add_widget(dlg)
    dlg.show()
    qttester.send_keypress(dlg, constants.Key.Key_F11)
    dlg.delete_on_close()
    btn = widgets.RadioButton("test")
    qtbot.addWidget(btn)
    dlg.box.add_widget(btn)
    dlg.set_icon("mdi.timer")
    dlg.resize(200, 400)
    dlg.resize((150, 400))
    dlg.add_buttonbox()


def test_dial(qtbot):
    widget = widgets.Dial()
    qtbot.addWidget(widget)


def test_dialogbuttonbox(qtbot):
    box = widgets.DialogButtonBox()
    qtbot.addWidget(box)
    box.set_horizontal()
    box.set_vertical()
    btn = box.add_default_button("apply")
    with pytest.raises(InvalidParamError):
        btn = box.add_default_button("test")
    box.set_orientation("horizontal")
    with pytest.raises(InvalidParamError):
        box.set_orientation("test")
    assert box.get_orientation() == "horizontal"
    box.add_button("test_dialogbuttonbox", callback=print)
    assert len(box) == 2
    assert btn == box["apply"]
    assert "apply" in box
    for _item in box:
        pass
    btn = box.add_default_buttons(["ok"])


def test_dockwidget(qtbot):
    widget = widgets.DockWidget()
    qtbot.addWidget(widget)
    widget.setup_title_bar()
    widget.maximize()
    w = widgets.Widget()
    qtbot.addWidget(w)
    widget.set_widget(w)
    w.raise_dock()


def test_doublespinbox(qtbot):
    widget = widgets.DoubleSpinBox(value=5)
    qtbot.addWidget(widget)
    widget.set_disabled()
    widget.set_enabled()


def test_filedialog(qtbot):
    dlg = widgets.FileDialog(path_id="test", extension_filter=dict(test=["*.test"]))
    qtbot.addWidget(dlg)
    dlg.set_label_text("accept", "test_filedialog")
    dlg.set_accept_mode("open")
    with pytest.raises(InvalidParamError):
        dlg.set_accept_mode("bla")
    dlg.set_view_mode("detail")
    with pytest.raises(InvalidParamError):
        dlg.set_view_mode("bla")
    dlg.set_label_text("filetype", "test")
    with pytest.raises(InvalidParamError):
        dlg.set_label_text("bla", "test")
    dlg.set_accept_mode("save")
    dlg.set_extension_filter(dict(a=[".csv"]))
    dlg.set_filter("all_dirs")
    with pytest.raises(InvalidParamError):
        dlg.set_filter("test")
    dlg.selected_file()
    dlg.selected_files()
    path = dlg.get_directory()
    dlg.set_directory(path)


# Segfault with PySide6
def test_filesystemmodel(qttester):
    model = widgets.FileSystemModel()
    model.set_root_path("/")
    idx = model.index(0, 0)
    model.get_paths([idx])
    model.data(idx, model.Roles.FilePathRole)
    model.watch_for_changes(False)
    model.use_custom_icons(False)
    model.resolve_sym_links(False)
    model.set_name_filters(["test"], hide=True)
    model.set_filter("drives")
    with pytest.raises(InvalidParamError):
        model.set_filter("test")


# modeltest.ModelTest(model)
# qttester.test_model(model, force_py=True)


def test_fontcombobox(qtbot):
    widget = widgets.FontComboBox()
    qtbot.addWidget(widget)
    font = widget.get_current_font()
    assert font == widget.get_value()
    widget.set_font_filters("scalable")
    assert widget.get_font_filters() == ["scalable"]
    with pytest.raises(InvalidParamError):
        widget.set_font_filters("test")


def test_fontdialog(qtbot):
    dlg = widgets.FontDialog()
    qtbot.addWidget(dlg)
    dlg.get_current_font()


def test_formlayout(qtbot):
    layout = widgets.FormLayout()
    layout.set_size_constraint("maximum")
    with pytest.raises(InvalidParamError):
        layout.set_size_constraint("bla")
    layout[0, "left"] = "0, left"
    button_1 = widgets.RadioButton("1, left")
    layout[1, "left"] = button_1
    layout[0, "right"] = "label 1 right"
    button_2 = widgets.RadioButton("1, right")
    layout[1, "right"] = button_2
    layout[2] = "by str"
    button_3 = widgets.RadioButton("widget[3]")
    layout[3] = button_3
    button_4 = widgets.RadioButton("added with +=")
    layout += button_4
    button_5 = widgets.RadioButton("tuple")
    layout += ("added with +=", button_5)
    assert len(layout) == 6
    layout.set_row_wrap_policy("wrap_long")
    assert layout.get_row_wrap_policy() == "wrap_long"
    with pytest.raises(InvalidParamError):
        layout.set_row_wrap_policy("test")
    layout.set_field_growth_policy("expanding_fields_grow")
    assert layout.get_field_growth_policy() == "expanding_fields_grow"
    with pytest.raises(InvalidParamError):
        layout.set_field_growth_policy("test")


def test_frame(qtbot):
    frame = widgets.Frame()
    frame.set_frame_shadow("raised")
    assert frame.get_frame_shadow() == "raised"
    with pytest.raises(InvalidParamError):
        frame.set_frame_shadow("test")
    frame.set_frame_shape("panel")
    assert frame.get_frame_shape() == "panel"
    with pytest.raises(InvalidParamError):
        frame.set_frame_shape("test")


def test_gesture():
    gesture = widgets.Gesture()
    assert gesture.get_state() == "none"
    assert gesture.get_gesture_type() == "custom"
    gesture.get_hot_spot()
    gesture.set_gesture_cancel_policy("all_in_context")
    with pytest.raises(InvalidParamError):
        gesture.set_gesture_cancel_policy("test")
    assert gesture.get_gesture_cancel_policy() == "all_in_context"


def test_graphicsblureffect():
    effect = widgets.GraphicsBlurEffect()
    effect.set_blur_hints("animation")
    assert effect.get_blur_hints() == ["animation"]


def test_graphicsitem(qtbot):
    item = widgets.GraphicsItem()
    item.set_panel_modality("scene")
    assert item.get_panel_modality() == "scene"
    with pytest.raises(InvalidParamError):
        item.set_panel_modality("test")
    item.set_cache_mode("item_coordinate")
    assert item.get_cache_mode() == "item_coordinate"
    with pytest.raises(InvalidParamError):
        item.set_cache_mode("test")
    item.set_focus("active_window")
    with pytest.raises(InvalidParamError):
        item.set_focus("test")
    item[0] = "test"
    assert item[0] == "test"
    # item.get_shape()


def test_graphicsgridlayout(qtbot):
    layout = widgets.GraphicsGridLayout()
    item = widgets.GraphicsProxyWidget()
    button_1 = widgets.RadioButton("Test")
    qtbot.addWidget(button_1)
    item.setWidget(button_1)
    item2 = widgets.GraphicsProxyWidget()
    button_2 = widgets.RadioButton("Test")
    qtbot.addWidget(button_2)
    item2.setWidget(button_2)
    layout[1, 5:6] = item
    layout += item2
    layout.set_column_alignment(0, "left")
    with pytest.raises(InvalidParamError):
        layout.set_column_alignment(0, "test")
    layout.set_row_alignment(0, "left")
    with pytest.raises(InvalidParamError):
        layout.set_row_alignment(0, "test")
    assert len(layout) == 2
    layout.set_margin(0)


def test_graphicspixmapitem(qtbot):
    item = widgets.GraphicsPixmapItem()
    item.set_transformation_mode("smooth")
    assert item.get_transformation_mode() == "smooth"
    with pytest.raises(InvalidParamError):
        item.set_transformation_mode("test")
    item.set_shape_mode("bounding_rect")
    assert item.get_shape_mode() == "bounding_rect"
    with pytest.raises(InvalidParamError):
        item.set_shape_mode("test")
    assert item.get_pixmap() is None


def test_graphicsscene(qtbot):
    scene = widgets.GraphicsScene()
    icon = iconprovider.get_icon("mdi.help-circle-outline")
    pixmap = icon.pixmap(200, 200)
    pixmap2 = icon.pixmap(20, 20)
    # item = widgets.GraphicsItem()
    g_1 = scene.add(pixmap)
    assert scene[0] == g_1
    g_2 = scene.add(pixmap2)
    assert scene.colliding_items(g_1, mode="intersects_bounding_rect") == [g_2]
    with pytest.raises(InvalidParamError):
        scene.colliding_items(g_1, mode="test")
    scene.add_line(core.LineF(0, 0, 1, 1))
    scene.add_line(core.Line(0, 0, 1, 1))
    scene.add_line((0, 0, 1, 1))
    scene.add_rect(core.Rect(0, 0, 1, 1))
    scene.add_rect(core.RectF(0, 0, 1, 1))
    scene.add_rect((0, 0, 1, 1), brush=gui.Brush(), pen=gui.Pen())
    scene.add_ellipse(core.Rect(0, 0, 1, 1))
    scene.add_ellipse(core.RectF(0, 0, 1, 1))
    scene.add_ellipse((0, 0, 1, 1), brush=gui.Brush(), pen=gui.Pen())
    poly = gui.PolygonF()
    poly.add_points((0, 0), (2, 0), (2, 1), (0, 1))
    scene.add_polygon(poly, brush=gui.Brush(), pen=gui.Pen())

    poly = gui.PolygonF()
    poly.add_points((0, 0), (2, 0), (2, 1), (0, 1))
    scene.add_polygon(poly)
    scene.add_pixmap(gui.Pixmap())
    path = gui.PainterPath()
    rect = core.RectF(0, 0, 1, 1)
    path.addRect(rect)
    scene.add_path(path, brush=gui.Brush(), pen=gui.Pen())
    scene.add_text("test", font=gui.Font())
    scene.add_simple_text("test")
    widget = widgets.Widget()
    scene.add_widget(widget)
    item = widgets.GraphicsRectItem(0, 0, 10, 10)
    scene.add_item_group(item)
    scene.set_item_index_method("none")
    assert scene.get_item_index_method() == "none"
    with pytest.raises(InvalidParamError):
        scene.set_item_index_method("test")


def test_graphicsrotation():
    rotation = widgets.GraphicsRotation()
    rotation.set_axis("z")
    with pytest.raises(InvalidParamError):
        rotation.set_axis("test")


def test_graphicsview(qtbot):
    view = widgets.GraphicsView()
    qtbot.addWidget(view)
    view.set_transformation_anchor("view_center")
    assert view.get_transformation_anchor() == "view_center"
    with pytest.raises(InvalidParamError):
        view.set_transformation_anchor("test")
    view.set_resize_anchor("view_center")
    assert view.get_resize_anchor() == "view_center"
    with pytest.raises(InvalidParamError):
        view.set_resize_anchor("test")
    view.set_viewport_update_mode("minimal")
    assert view.get_viewport_update_mode() == "minimal"
    with pytest.raises(InvalidParamError):
        view.set_viewport_update_mode("test")
    view.set_drag_mode("scroll_hand")
    assert view.get_drag_mode() == "scroll_hand"
    with pytest.raises(InvalidParamError):
        view.set_drag_mode("test")
    view.set_rubberband_selection_mode("intersects_shape")
    assert view.get_rubberband_selection_mode() == "intersects_shape"
    with pytest.raises(InvalidParamError):
        view.set_rubberband_selection_mode("test")
    view.set_cache_mode("background")
    assert view.get_cache_mode() == "background"
    with pytest.raises(InvalidParamError):
        view.set_cache_mode("test")


def test_gridlayout(qtbot):
    layout = widgets.GridLayout()
    layout2 = widgets.GridLayout()
    widget = widgets.RadioButton()
    layout[0:1, 0:3] = widget
    layout[5, 5] = layout2
    assert layout[0, 0] == widget
    assert layout[5, 5] == layout2
    layout.set_size_constraint("maximum")
    layout.set_alignment("left")
    assert len(layout) == len(list(layout)) == 2
    layout += widgets.RadioButton()
    layout.set_origin_corner("bottom_right")
    assert layout.get_origin_corner() == "bottom_right"
    with pytest.raises(InvalidParamError):
        layout.set_origin_corner("test")


def test_groupbox(qtbot):
    widget = widgets.GroupBox()
    qtbot.addWidget(widget)
    widget.set_title("test_groupbox")
    ly = widgets.HBoxLayout()
    widget.set_layout(ly)
    ly += widgets.RadioButton("+=")
    widget.set_alignment("left")
    widget.set_enabled(False)


@pytest.mark.skipif(sys.platform == "linux", reason="X11 connection break")
def test_headerview(qtbot):
    def test():
        pass

    table = widgets.TableView()
    qtbot.addWidget(table)
    model = gui.StandardItemModel.from_dict(dict(a=[1, 2, 3]))
    table.set_model(model)
    header = widgets.HeaderView("horizontal", parent=table)
    table.setHorizontalHeader(header)
    header.set_resize_mode("interactive")
    header[0].set_resize_mode("interactive")
    with pytest.raises(InvalidParamError):
        header.set_resize_mode("test")
    header.resize_sections("interactive")
    header.set_context_menu_policy("custom")
    header.set_default_section_size(None)
    header.set_default_section_size(30)
    header.stretch_last_section()
    with pytest.raises(InvalidParamError):
        header.set_context_menu_policy("test")
    assert header.get_context_menu_policy() == "custom"
    header.set_custom_menu(test)
    assert len(header.get_section_labels()) == 1
    header.save_state()
    header.load_state()
    header.set_section_hidden(0, True)


def test_inputdialog(qtbot):
    dlg = widgets.InputDialog()
    qtbot.addWidget(dlg)
    dlg.set_input_mode("double")
    with pytest.raises(InvalidParamError):
        dlg.set_input_mode("test")
    assert dlg.get_input_mode() == "double"
    dlg.set_text_echo_mode("no_echo")
    with pytest.raises(InvalidParamError):
        dlg.set_text_echo_mode("test")
    assert dlg.get_text_echo_mode() == "no_echo"


def test_keysequenceedit(qtbot):
    seq = gui.KeySequence("Ctrl+A")
    edit = widgets.KeySequenceEdit(seq)
    edit.set_value("Ctrl+A")
    assert edit.get_value() == "Ctrl+A"
    assert edit.is_valid()


# not workin on Mac + PySide6
# def test_label(qtbot):
#     label = widgets.Label()
#     qtbot.addWidget(label)
#     label.set_image("")
#     label.set_text("test_label")
#     label.set_bold()
#     label.set_italic()
#     label.set_indent(4)
#     label.set_weight("extra_light")
#     label.set_point_size(14)
#     label.set_color("red")
#     label.set_color(None)
#     with pytest.raises(InvalidParamError):
#         label.set_weight("test")
#     with pytest.raises(InvalidParamError):
#         label.set_text_format("test")
#     label.set_alignment(horizontal="left", vertical="top")
#     label.set_alignment(vertical="bottom")
#     label.set_alignment()
#     label.set_text_interaction("by_mouse")
#     expected = ["by_mouse", "like_text_editor", "like_text_browser"]
#     assert label.get_text_interaction() == expected
#     label.allow_links()
#     with pytest.raises(InvalidParamError):
#         label.set_text_interaction("test")
# assert label.get_text_interaction() == "by_mouse"


def test_layoutitem(qtbot):
    item = widgets.LayoutItem()
    assert item.get_item() is None
    item.set_alignment("right")
    assert item.get_alignment() == "right"


def test_lcdnumber(qtbot):
    lcd = widgets.LCDNumber()
    qtbot.addWidget(lcd)
    lcd.set_value(500)
    assert lcd.get_value() == 500
    lcd.set_segment_style("filled")
    with pytest.raises(InvalidParamError):
        lcd.set_segment_style("test")
    assert lcd.get_segment_style() == "filled"
    lcd.set_mode("octal")
    with pytest.raises(InvalidParamError):
        lcd.set_mode("test")
    assert lcd.get_mode() == "octal"


def test_lineedit(qtbot):
    widget = widgets.LineEdit("test_lineedit")
    qtbot.addWidget(widget)
    widget.set_regex_validator("[0-9]")
    widget.set_font("Consolas")
    widget.set_text("0")
    widget.append_text("a")
    widget.set_echo_mode("password")
    with pytest.raises(InvalidParamError):
        widget.set_echo_mode("test")
    assert widget.get_echo_mode() == "password"
    widget.set_cursor_move_style("visual")
    with pytest.raises(InvalidParamError):
        widget.set_cursor_move_style("test")
    assert widget.get_cursor_move_style() == "visual"
    widget.set_input_mask("X")
    widget.set_range(0, 10)
    widget.set_value("5")
    widget += "append"


def test_listview(qtbot):
    widget = widgets.ListView()
    qtbot.addWidget(widget)
    widget.set_selection_mode(None)
    widget.set_selection_mode("single")
    widget.toggle_select_all()
    widget.set_selection_mode("multi")
    assert widget.get_selection_mode() == "multi"
    widget.set_view_mode("icon")
    with pytest.raises(InvalidParamError):
        widget.set_view_mode("test")
    assert widget.get_view_mode() == "icon"


def test_listwidget(qtbot):
    widget = widgets.ListWidget()
    qtbot.addWidget(widget)
    widget.add("test_listwidget", icon="mdi.timer")
    widget.add("test_listwidget", icon="mdi.timer")
    assert len(widget) == 2
    item = widget[0]
    assert item is not None
    widget.scroll_to_item(item, mode="ensure_visible")
    with pytest.raises(InvalidParamError):
        widget.scroll_to_item(item, mode="test")
    widget.on_index_change()
    widget.add_items(["a", "b"])
    widget.set_value("b")
    widget.sort()
    for _item in widget:
        pass
    item = widgets.ListWidgetItem()
    widget += item
    del widget[0]
    widget.add_item(
        "Test",
        icon="mdi.timer",
        data={1: "Test"},
        foreground=gui.Brush(),
        background=gui.Brush(),
        font=gui.Font(),
        selected=True,
        status_tip="test",
        tool_tip="test",
        whats_this="test",
        checkstate="unchecked",
        size_hint=core.Size(10, 10),
        is_user_type=True,
    )


def test_listwidgetitem(qtbot):
    item = widgets.ListWidgetItem()
    item[constants.USER_ROLE] = "test"
    assert item[constants.USER_ROLE] == "test"
    item.set_icon("mdi.timer")
    item.set_checkstate("unchecked")
    with pytest.raises(InvalidParamError):
        item.set_checkstate("test")
    assert item.get_checkstate() == "unchecked"
    item.get_background()
    item.get_foreground()
    item.get_font()
    item.get_icon()
    bytes(item)


def test_mainwindow(qtbot):
    window = widgets.MainWindow()
    qtbot.addWidget(window)
    window.set_icon("mdi.timer")
    window.setObjectName("mainwindow")
    dockwidget = widgets.DockWidget()
    qtbot.addWidget(dockwidget)
    dockwidget.setObjectName("dockwidget")
    window.add_dockwidget(dockwidget, "left")
    widget = widgets.MainWindow()
    qtbot.addWidget(widget)
    widget.setObjectName("test")
    window.set_widget(widget)
    assert window["test"] == widget
    window.show()
    window.close()
    window.save_window_state(recursive=True)
    window.load_window_state(recursive=True)
    window.toggle_fullscreen()
    window.toggle_fullscreen()
    toolbar = widgets.ToolBar()
    qtbot.addWidget(toolbar)
    window.add_toolbar(toolbar)
    with pytest.raises(InvalidParamError):
        toolbar_2 = widgets.ToolBar()
        window.add_toolbar(toolbar_2, area="test")
    assert len(window.get_toolbars()) == 1
    assert len(window.get_docks()) == 1
    window.remove(dockwidget)
    window.add_toolbar_break()
    with pytest.raises(InvalidParamError):
        window.add_toolbar_break("test")
    with window.edit_stylesheet() as ss:
        ss.QMainWindow.separator.setValues(width="1px", border="none")
    window.createPopupMenu()
    window.add_widget_as_dock("test", "Title")
    window.set_icon(None)


def test_mdiarea(qtbot):
    area = widgets.MdiArea()
    qtbot.addWidget(area)
    area.set_window_order("activation_history")
    with pytest.raises(InvalidParamError):
        area.set_window_order("test")
    assert area.get_window_order() == "activation_history"
    area.set_view_mode("default")
    with pytest.raises(InvalidParamError):
        area.set_view_mode("test")
    assert area.get_view_mode() == "default"
    area.set_tab_position("north")
    with pytest.raises(InvalidParamError):
        area.set_tab_position("test")
    assert area.get_tab_position() == "north"
    area.set_background("black")
    widget_1 = widgets.Widget()
    qtbot.addWidget(widget_1)
    area.add(widget_1)
    widget_2 = widgets.Widget()
    qtbot.addWidget(widget_2)
    area += widget_2
    sub = widgets.MdiSubWindow()
    qtbot.addWidget(sub)
    area.add(sub)


def test_menu(qtbot):
    menu = widgets.Menu("1", icon="mdi.timer")
    menu.add(gui.Action(text="TestAction"))
    act = gui.Action(text="TestAction")
    act.setObjectName("test")
    menu += act
    assert menu["test"] == act
    with pytest.raises(KeyError):
        menu["bla"]

    def test():
        pass

    menu.add_action(
        "test_menu",
        triggered=test,
        icon="mdi.timer",
        shortcut="Ctrl+A",
        checkable=True,
        status_tip="test",
    )
    assert len(menu) == 3
    for _item in menu:
        pass
    menu.add_menu(widgets.Menu())
    menu.add_separator("test_menu")
    menu.add_separator()


def test_menubar(qtbot):
    menu = widgets.MenuBar()
    qtbot.addWidget(menu)
    menu += gui.Action(text="TestAction")
    menu += widgets.Menu("TestMenu")
    menu.add_action(gui.Action(text="TestAction 2"))
    menu.add_menu(widgets.Menu("TestMenu 2"))
    menu.add_separator()
    menu.add_action("test_menubar")
    menu.add_menu("test_menubar")


def test_messagebox(qtbot):
    widget = widgets.MessageBox(buttons=["reset"])
    widget.set_icon("warning")
    widget.set_icon("mdi.timer")
    widget.add_button("ok")
    widget.set_text_format("rich")
    with pytest.raises(InvalidParamError):
        widget.set_text_format("test")
    assert widget.get_text_format() == "rich"
    with pytest.raises(InvalidParamError):
        widget.add_button("test")
    widget.get_standard_buttons()


def test_pangesture():
    gesture = widgets.PanGesture()
    gesture.get_delta()
    gesture.get_last_offset()
    gesture.get_offset()


def test_plaintextdocumentlayout():
    doc = gui.TextDocument()
    layout = widgets.PlainTextDocumentLayout(doc)
    repr(layout)
    assert len(layout) == 1
    layout.get_block_bounding_rect(gui.TextBlock())
    layout.get_frame_bounding_rect(gui.TextFrame(doc))


def test_plaintextedit(qtbot):
    widget = widgets.PlainTextEdit("This is a test")
    qtbot.addWidget(widget)
    with widget.selecter.create_cursor() as c:
        c.select_text(2, 4)
    with widget.selecter.current_cursor() as c:
        c.select_text(2, 4)
    widget.selecter.select_text(2, 4)
    widget.set_text("hallo")
    widget.set_disabled()
    widget.allow_wheel_zoom()
    widget.set_enabled()
    widget.set_font("Consolas")
    widget.append_text(" test")
    widget.append_text("test", newline=False)
    assert widget.text() == "hallo\n testtest"
    widget.selecter.highlight_current_line()
    widget.set_read_only()
    widget.scroll_to_top()
    widget.scroll_to_bottom()
    widget.set_value("test")
    widget.set_word_wrap_mode("anywhere")
    with pytest.raises(InvalidParamError):
        widget.set_word_wrap_mode("test")
    assert widget.get_word_wrap_mode() == "anywhere"
    widget.set_line_wrap_mode("widget_width")
    with pytest.raises(InvalidParamError):
        widget.set_line_wrap_mode("test")
    assert widget.get_line_wrap_mode() == "widget_width"
    assert widget.get_value() == "test"
    widget += "append"
    widget.set_regex_validator("[0-9]")


def test_pinchgesture():
    gesture = widgets.PinchGesture()
    gesture.get_start_center_point()
    gesture.get_center_point()
    gesture.get_last_center_point()
    gesture.set_change_flags(center_point=True)
    assert gesture.get_change_flags() == ["center_point"]
    gesture.set_total_change_flags(center_point=True)
    assert gesture.get_total_change_flags() == ["center_point"]


def test_progressbar(qtbot):
    bar = widgets.ProgressBar()
    qtbot.addWidget(bar)
    bar.set_alignment("left")
    bar.set_alignment("right")
    assert bar.get_alignment() == "right"
    with pytest.raises(InvalidParamError):
        bar.set_alignment("test")
    # assert bar.get_alignment() == "left"
    bar.set_text_direction("top_to_bottom")
    # assert bar.get_text_direction() == "top_to_bottom"
    with pytest.raises(InvalidParamError):
        bar.set_text_direction("test")
    bar.set_range(0, 20)
    # assert bar.get_text_direction() == "top_to_bottom"


# def test_progressdialog(qtbot):
#     widgets.ProgressDialog()


def test_pushbutton(qtbot):
    widget = widgets.PushButton("test_pushbutton", clicked=print)
    qtbot.addWidget(widget)
    widget.set_text("test_pushbutton")
    widget.set_disabled()
    widget.set_enabled()
    assert widget.get_value() is False
    widget.set_icon("mdi.timer")
    widget.get_icon()
    widget.set_style_icon("titlebar_close_button")
    widget.set_icon_size(10)
    assert widget.get_icon_size() == core.Size(10, 10)
    widget.set_value(True)


def test_radiobutton(qtbot):
    widget = widgets.RadioButton("test_radiobutton")
    qtbot.addWidget(widget)
    widget.set_icon("mdi.timer")
    widget.set_enabled()
    widget.set_disabled()
    assert not bool(widget)
    widget.set_value(True)
    assert widget.get_value() is True


def test_rubberband(qtbot):
    band = widgets.RubberBand("line")
    qtbot.addWidget(band)
    assert band.get_shape() == "line"


def test_scrollarea(qtbot):
    widget = widgets.ScrollArea()
    qtbot.addWidget(widget)
    widget.set_widget(widgets.Widget())


def test_scrollerproperties():
    properties = widgets.ScrollerProperties()
    properties["snap_time"] = 100
    with pytest.raises(InvalidParamError):
        properties.set_scroll_metric("test", 100)
    assert properties["snap_time"] == 100
    with pytest.raises(InvalidParamError):
        properties.get_scroll_metric("test")


def test_scroller(qtbot):
    w = widgets.PlainTextEdit()
    qtbot.addWidget(w)
    scroller = widgets.Scroller.get_scroller(w)
    assert scroller.get_state() == "inactive"
    scroller.get_velocity()
    scroller.get_pixel_per_meter()
    scroller.get_final_position()
    scroller.handle_input("move", core.PointF())
    with pytest.raises(InvalidParamError):
        scroller.handle_input("test", core.PointF())
    scroller.get_scroller_properties()
    # weirdly this returns "tap" instead of "tap_and_hold" when PDFView class is removed.
    assert widgets.Scroller.grab_gesture(w) in ["tap_and_hold", "tap"]


def test_sizepolicy(qtbot):
    pol = widgets.SizePolicy()
    pol.set_control_type("toolbutton")
    assert pol.get_control_type() == "toolbutton"


def test_slider(qtbot):
    widget = widgets.Slider()
    qtbot.addWidget(widget)
    widget.set_horizontal()
    assert widget.is_horizontal()
    widget.set_vertical()
    assert widget.is_vertical()
    with pytest.raises(InvalidParamError):
        widget.set_repeat_action("test")
    widget.set_repeat_action("page_step_add")
    assert widget.get_repeat_action() == "page_step_add"
    with pytest.raises(InvalidParamError):
        widget.set_tick_position("test")
    widget.set_tick_position("below")
    assert widget.get_tick_position() == "below"
    widget.set_orientation("horizontal")
    with pytest.raises(InvalidParamError):
        widget.set_orientation("test")
    assert widget.get_orientation() == "horizontal"
    widget.set_auto_scroll_to_end(True)
    widget.set_auto_scroll_to_end(False)


def test_statusbar(qtbot):
    widget = widgets.MainWindow()
    qtbot.addWidget(widget)
    status_bar = widgets.StatusBar()
    qtbot.addWidget(status_bar)
    label = widgets.Label("test_statusbar")
    qtbot.addWidget(label)
    status_bar.addWidget(label)
    status_bar.show_message("test_statusbar")
    status_bar.add_action(gui.Action())
    status_bar += gui.Action()
    widget.setStatusBar(status_bar)
    status_bar.add_widget(widgets.Widget())
    status_bar += widgets.Widget()
    status_bar.add_widget(widgets.Widget(), permanent=True)


def test_stackedlayout(qtbot):
    layout = widgets.StackedLayout()
    widget = widgets.RadioButton("test_stackedlayout")
    qtbot.addWidget(widget)
    layout += widget
    assert widget in layout
    layout.set_current_widget(widget)
    layout.set_size_constraint("maximum")
    layout.set_margin(0)
    assert len(layout) == 1
    for _item in layout:
        pass


def test_stackedwidget(qtbot):
    widget = widgets.StackedWidget()
    w = widgets.RadioButton("test_stackedwidget")
    widget += w
    assert widget[0] == w
    assert w in widget
    widget.set_current_widget(w)
    assert len(widget) == 1
    for _item in widget:
        pass


def test_spaceritem(qtbot):
    item = widgets.SpacerItem(0, 0, "expanding", "expanding")
    item.change_size(0, 0)


def test_spinbox(qtbot):
    widget = widgets.SpinBox(value=5)
    qtbot.addWidget(widget)
    widget.set_disabled()
    widget.set_enabled()
    widget.set_value(10)
    widget.set_special_value("test_spinbox")
    with pytest.raises(InvalidParamError):
        widget.set_button_symbols("test")
    with pytest.raises(InvalidParamError):
        widget.set_correction_mode("test")
    with pytest.raises(InvalidParamError):
        widget.set_step_type("test")
    assert widget.is_valid()
    assert widget.get_value() == 10


def test_splashscreen(qtbot):
    pixmap = gui.Pixmap.create_dot()
    scr = widgets.SplashScreen(pixmap)
    qtbot.addWidget(scr)
    with scr:
        pass
    scr.show_message("test")


def test_splitter(qtbot):
    widget = widgets.Splitter("vertical")
    qtbot.addWidget(widget)
    test = widgets.Label("test_splitter")
    qtbot.addWidget(widget)
    test2 = widgets.Label("test_splitter")
    qtbot.addWidget(widget)
    widget.add_widget(test)
    widget += test2
    assert len(widget) == 2
    assert widget[0] == test
    for _item in widget:
        pass
    widget.set_size_policy("expanding", "expanding")
    widget.set_orientation("horizontal")
    with pytest.raises(InvalidParamError):
        widget.set_orientation("test")
    widget.add_layout(widgets.HBoxLayout())
    widgets.Splitter.from_widgets(widgets.Widget())
    test3 = widgets.Label("test_splitter")
    widget[0] = test3


def test_styleoptionslider(qtbot):
    slider = widgets.StyleOptionSlider()
    slider.is_vertical()
    slider.set_horizontal()
    slider.set_vertical()


def test_swipegesture():
    gesture = widgets.SwipeGesture()
    assert gesture.get_horizontal_direction() == "right"
    assert gesture.get_vertical_direction() == "none"


def test_systemtrayicon(qtbot):
    icon = widgets.SystemTrayIcon()
    icon.set_icon("mdi.folder")
    icon.show_message("test", "", "critical")
    icon.show_message("test", "", "mdi.folder")
    icon.show_message("test", "")


@pytest.mark.skipif(sys.platform == "linux", reason="X11 connection break")
def test_tabwidget(qtbot):
    widget = widgets.TabWidget(detachable=True)
    qtbot.addWidget(widget)
    widget1 = widgets.Widget()
    qtbot.addWidget(widget1)
    # adding tabs seems to hang PySide6, need to investigate.
    # widget.add_tab(widget1, "mdi.timer", show=True)
    widget.set_document_mode(True)
    widget2 = widgets.Widget()
    qtbot.addWidget(widget2)
    # widget.add_tab(widget2, "test_tabwidget", "mdi.timer", position=0, show=True)
    #     assert len(widget) == 2
    #     w = widgets.Widget()
    #     widget.add_tab(w, "test_tabwidget", "mdi.timer")
    #     assert widget[2] == w
    #     widget.set_tab(0, "right", None)
    widget.set_detachable()
    #     # widget.detach_tab(0, core.Point())
    with pytest.raises(InvalidParamError):
        widget.set_tab_shape("test")
    #     widget.remove_tab(1)
    # layout = widgets.HBoxLayout()
    #     widget.add_tab(layout, "mdi.timer")
    widget.set_icon_size(10)
    widget.set_icon_size((10, 10))
    widget.set_icon_size(core.Size(10, 10))
    widget.close_detached_tabs()
    # widget.close_detached_tabs()


def test_tapgesture():
    gesture = widgets.TapGesture()
    gesture.get_position()


def test_tapandholdgesture():
    gesture = widgets.TapAndHoldGesture()
    gesture.get_position()


def test_textbrowser(qtbot):
    tmp = tempfile.NamedTemporaryFile(delete=False)
    tmp.write(b"Test")
    tmp.close()
    path = pathlib.Path(tempfile.gettempdir()) / tmp.name
    widget = widgets.TextBrowser()
    widget.set_markdown("test")
    widget.set_markdown_file(str(path))
    # widget.set_rst("test")
    # widget.set_rst_file(str(path))
    path.unlink()


def test_textedit(qtbot):
    widget = widgets.TextEdit()
    qtbot.addWidget(widget)
    widget.set_text("test")
    widget.append_text(" this")
    assert widget.text() == "test\n this"
    with widget.selecter.create_cursor() as c:
        c.select_text(1, 3)
    widget.selecter.select_text(1, 3)
    widget.set_font("Consolas")
    widget.set_enabled()
    widget.set_read_only()
    widget.scroll_to_bottom()
    widget.set_disabled()
    widget.set_background_color("black")
    widget.set_text_color("red")
    widget += "append"


def test_timeedit(qtbot):
    widget = widgets.TimeEdit()
    qtbot.addWidget(widget)
    widget.set_disabled()
    widget.set_enabled()
    widget.set_range(datetime.time(1, 1, 1), datetime.time(3, 3, 3))
    widget.set_value(datetime.time(0, 0, 0))
    assert widget.get_time() == widget.min_time()
    assert widget.max_time() == datetime.time(3, 3, 3)
    dt = datetime.time(2, 2, 2)
    widget.set_value(dt)
    assert widget.get_value() == dt


def test_toolbar(qtbot):
    widget = widgets.ToolBar()
    qtbot.addWidget(widget)
    widget.add_menu_button(label="test,", icon="mdi.timer", menu=widgets.Menu())
    widget.set_style("icon")
    widget.add_separator("Test")
    widget.add_separator()
    widget.add_spacer()
    assert widget.get_style() == "icon"
    widget.set_font_size(10)
    widget.set_enabled()
    widget.set_disabled()
    widget.set_icon_size(10)
    assert widget.get_icon_size() == core.Size(10, 10)
    assert widget.is_area_allowed("top")
    with pytest.raises(InvalidParamError):
        widget.is_area_allowed("test")
    assert widget.get_allowed_areas() == ["left", "right", "top", "bottom", "all"]

    def test():
        pass

    widget.add_action(text="test", icon="mdi.timer", triggered=test, checkable=True)


def test_toolbutton(qtbot):
    widget = widgets.ToolButton()
    qtbot.addWidget(widget)
    widget.set_disabled()
    widget.set_enabled()
    widget.set_shortcut(None)
    widget.set_shortcut("Ctrl+A")
    assert widget.get_shortcut()
    widget.set_icon_size((20, 10))
    widget.set_icon_size(20)
    action = gui.Action()
    widget.set_default_action(action)
    with pytest.raises(InvalidParamError):
        widget.set_popup_mode("test")
    assert widget.get_popup_mode() == "delayed"
    with pytest.raises(InvalidParamError):
        widget.set_arrow_type("test")
    widget.set_arrow_type("left")
    with pytest.raises(InvalidParamError):
        widget.set_style("test")
    widget.set_style("text_below_icon")
    assert widget.get_arrow_type() == "left"
    menu = widgets.Menu()
    act = gui.Action()
    act.setObjectName("test")
    menu.add(act)
    widget.setMenu(menu)
    assert widget["test"] == act


def test_tooltip(qtbot):
    widgets.ToolTip.show_text(text="test")


def test_tabbar(qtbot):
    widget = widgets.TabBar()
    qtbot.addWidget(widget)
    widget.set_icon_size(10)
    assert widget.get_icon_size() == core.Size(10, 10)
    with pytest.raises(InvalidParamError):
        widget.set_selection_behavior_on_remove("test")
    assert widget.get_remove_behaviour() == "left_tab"
    with pytest.raises(InvalidParamError):
        widget.set_elide_mode("test")


def test_tableview(qtbot):
    widget = widgets.TableView()
    qtbot.addWidget(widget)
    widget.set_selection_mode("extended")
    with pytest.raises(InvalidParamError):
        widget.set_selection_mode("test")
    widget.set_selection_behavior("rows")
    assert widget.get_selection_behavior() == "rows"
    with pytest.raises(InvalidParamError):
        widget.set_selection_behavior("test")
    widget.set_horizontal_scrollbar_policy("always_on")
    widget.set_vertical_scrollbar_policy("always_on")
    widget.set_horizontal_scrollbar_width(12)
    widget.set_vertical_scrollbar_width(12)
    widget.set_edit_triggers(None)
    widget.set_edit_triggers("edit_key_pressed")
    widget.sort_by_column(0)
    with pytest.raises(InvalidParamError):
        widget.set_edit_triggers("test")
    widget.selectAll()
    widget.current_index()
    widget.current_data()
    widget.adapt_sizes()
    widget.setup_list_style()
    widget.setup_dragdrop_move()
    widget.num_selected()
    widget.jump_to_column(0)
    # widget.select_last_row()
    widget.set_scroll_mode("item")
    widget.set_vertical_scroll_mode("item")
    widget.set_horizontal_scroll_mode("item")
    with pytest.raises(InvalidParamError):
        widget.set_scroll_mode("aa")
    with pytest.raises(InvalidParamError):
        widget.set_vertical_scroll_mode("aa")
    with pytest.raises(InvalidParamError):
        widget.set_horizontal_scroll_mode("aa")
    widget.highlight_when_inactive()
    widget.set_table_color("black")
    widget.scroll_to_bottom()
    widget.selected_data()
    widget.selected_rows()
    widget.selected_names()
    assert len(widget.selected_indexes()) == 0
    widget.get_edit_triggers()
    assert widget.h_header is not None
    header = widgets.HeaderView("vertical", parent=widget)
    qtbot.addWidget(header)
    widget.v_header = header
    assert widget.v_header is not None


def test_tablewidget(qtbot, tablewidget):
    # qtbot.addWidget(tablewidget)
    tablewidget.sort()
    item = widgets.TableWidgetItem("test")
    tablewidget[0, 0] = item
    assert tablewidget[0, 0] == item
    del tablewidget[0, 0]


def test_tablewidgetitem(qtbot):
    item = widgets.TableWidgetItem()
    item[constants.USER_ROLE] = "test"
    assert item[constants.USER_ROLE] == "test"
    item.set_icon("mdi.timer")
    item.set_checkstate("unchecked")
    with pytest.raises(InvalidParamError):
        item.set_checkstate("test")
    assert item.get_checkstate() == "unchecked"
    item.get_background()
    item.get_foreground()
    item.get_font()
    item.get_icon()


def test_tablewidgetselectionrange():
    range_1 = widgets.TableWidgetSelectionRange(5, 5, 15, 15)
    range_2 = widgets.TableWidgetSelectionRange(0, 0, 20, 20)
    result = widgets.TableWidgetSelectionRange(0, 0, 20, 20)
    assert result == range_1 | range_2
    result = widgets.TableWidgetSelectionRange(5, 5, 15, 15)
    assert result == range_1 & range_2
    assert repr(result) == "TableWidgetSelectionRange(5, 5, 15, 15)"


def test_toolbox(qtbot):
    w = widgets.RadioButton("test1")
    w2 = widgets.RadioButton("test2")
    qtbot.addWidget(w)
    qtbot.addWidget(w2)
    w2.setObjectName("test_name")
    widget = widgets.ToolBox()
    qtbot.addWidget(widget)
    widget.add_widget(w, "title", "mdi.timer")
    widget.add_widget(w2)
    assert widget["test_name"] == w2
    for _w in widget:
        pass
    assert widget[1] == w2
    del widget[1]


def test_treeview(qtbot):
    widget = widgets.TreeView()
    qtbot.addWidget(widget)
    assert len(widget) == 0
    model = gui.StandardItemModel.from_dict(dict(a=[1, 2, 3], b=["d", "e", "f"]))
    widget.set_model(model)
    widget.selectAll()
    assert widget.v_scrollbar
    widget.set_size_adjust_policy("content")
    with pytest.raises(InvalidParamError):
        widget.set_size_adjust_policy("test")
    assert widget.get_size_adjust_policy() == "content"
    widget.set_scrollbar_policy("always_on")
    with pytest.raises(InvalidParamError):
        widget.set_scrollbar_policy("test")
    widget.set_scrollbar_width(10)
    widget.setup_list_style()
    widget.setup_dragdrop_move()
    widget.sort_by_column(0)
    widget.scroll_to_top()
    widget.current_index()
    widget.set_selection_mode("extended")
    widget.set_drag_drop_mode("drop")
    widget.set_selection_behavior("rows")
    widget.set_horizontal_scrollbar_policy("always_on")
    widget.set_vertical_scrollbar_policy("always_on")
    with pytest.raises(InvalidParamError):
        widget.set_horizontal_scrollbar_policy("test")
    with pytest.raises(InvalidParamError):
        widget.set_vertical_scrollbar_policy("test")
    widget.set_horizontal_scrollbar_width(12)
    widget.set_vertical_scrollbar_width(12)
    widget.num_selected()
    widget.jump_to_column(0)
    widget.highlight_when_inactive()
    widget.raise_dock()
    widget.adapt_sizes()
    model = gui.StandardItemModel()
    model.add("test")
    widget.set_model(model)
    widget.set_delegate(widgets.StyledItemDelegate())
    widget.set_delegate(widgets.StyledItemDelegate(), column=0, persistent=True)
    widget.set_delegate(widgets.StyledItemDelegate(), row=0, persistent=True)
    widget.toggle_select_all()
    widget.toggle_select_all()
    model.update_row(0)


def test_treewidget(qtbot):
    widget = widgets.TreeWidget()
    qtbot.addWidget(widget)
    widget.sort()


def test_treewidgetitem(qtbot):
    item = widgets.TreeWidgetItem()
    item.set_icon("mdi.timer")
    item.set_checkstate("unchecked")
    with pytest.raises(InvalidParamError):
        item.set_checkstate("test")
    assert item.get_checkstate() == "unchecked"
    item.get_background()
    item.get_foreground()
    item.get_font()
    item.get_icon()
    item.set_child_indicator_policy("dont_show")
    with pytest.raises(InvalidParamError):
        item.set_child_indicator_policy("test")
    assert item.get_child_indicator_policy() == "dont_show"
    bytes(item)
    child = widgets.TreeWidgetItem()
    item += child
    for _child in item:
        pass
    assert len(item) == 1
    assert item[0] == child
    item.sort_children(0)
    del item[0]


def test_treewidgetitemiterator(qtbot):
    item = widgets.TreeWidget()
    widgets.TreeWidgetItemIterator(
        item,
        flags=None,
        hidden=True,
        selected=True,
        selectable=True,
        draggable=True,
        droppable=True,
        has_children=True,
        checked=True,
        enabled=True,
        editable=True,
        user_flag=True,
    )


def test_undoview(qtbot):
    view = widgets.UndoView()
    stack = gui.UndoStack()
    cmd = stack.add_command("test", redo=lambda: print("a"), undo=lambda: print("b"))
    view.set_clean_icon("mdi.folder")
    view.set_value(stack)
    assert view[0] == cmd


def test_whatsthis(qtbot):
    with widgets.WhatsThis.enter_mode():
        pass


def test_widget(qtbot):
    widget = widgets.Widget()
    qtbot.addWidget(widget)
    widget.set_tooltip("test")
    widget.set_cursor("caret")
    widget.set_min_width(100)
    widget.set_min_width(None)
    widget.set_max_width(100)
    widget.set_max_width(None)
    widget.set_min_height(200)
    widget.set_min_height(None)
    widget.set_max_height(200)
    widget.set_max_height(None)
    widget.set_font_size(20)
    widget.get_font_metrics()
    widget.get_font_info()
    widget.setObjectName("test")
    widget.set_unique_id()
    widget.get_palette()
    widget.set_attributes(native_window=False)
    with pytest.raises(InvalidParamError):
        widget.set_cursor("test")
    widget.set_focus_policy("strong")
    assert widget.get_focus_policy() == "strong"
    with pytest.raises(InvalidParamError):
        widget.set_focus_policy("test")
    layout = widgets.HBoxLayout()
    widget.set_layout(layout)
    with pytest.raises(ValueError):
        widget.set_layout("test")
    with widget.signals_blocked():
        pass
    with widget.edit_font():
        pass
    with widget.edit_stylesheet():
        pass
    with widget.edit_palette():
        pass
    widget.get_font()
    widget.set_enabled()
    widget.set_disabled()
    widget.set_min_size(1, 1)
    widget.set_max_size(2, 2)
    widget.set_title("test")
    assert widget.get_title() == "test"
    with widget.updates_off():
        widget.set_title("test2")
    widget.enabled = True
    assert widget.enabled
    widget.set_modality("window")
    with pytest.raises(InvalidParamError):
        widget.set_modality("test")
    assert widget.get_modality() == "window"
    widget.position_on("screen")
    widget_2 = widgets.Widget()
    qtbot.addWidget(widget_2)
    widget_2.set_layout("horizontal", margin=2)
    widget_3 = widgets.Widget()
    qtbot.addWidget(widget_3)
    widget_3.set_layout("form")
    widget_4 = widgets.Widget()
    qtbot.addWidget(widget_4)
    widget_4.set_layout("stacked")
    widget_5 = widgets.Widget()
    qtbot.addWidget(widget_5)
    widget_5.set_layout("flow")
    widget_5.set_margin(2)
    widget_5.set_window_state("fullscreen")
    widget_5.set_mask((0, 0, 10, 10), "ellipse")
    with pytest.raises(InvalidParamError):
        widget_5.set_window_state("test")
    assert widget_5.get_window_state() == "fullscreen"


def test_widgetaction(qtbot):
    action = gui.Action()
    widgetaction = widgets.WidgetAction(action)
    widgetaction.set_tooltip("test")
    widgetaction.set_enabled()
    widgetaction.set_disabled()
    widgetaction.set_icon("mdi.timer")
    widgetaction.set_shortcut("Ctrl+A")


def test_wizard(qtbot):
    w = widgets.Wizard()
    qtbot.addWidget(w)
    w.add_widget_as_page(widgets.Widget())
    pix = gui.Pixmap(100, 100)
    w.set_pixmap("background", pix)
    with pytest.raises(InvalidParamError):
        w.set_pixmap("test", pix)
    assert bytes(w.get_pixmap("background")) == bytes(pix)
    with pytest.raises(InvalidParamError):
        w.get_pixmap("test")
    w.set_wizard_style("mac")
    with pytest.raises(InvalidParamError):
        w.set_wizard_style("test")
    assert w.get_wizard_style() == "mac"
    w.set_button_text("back", "test")
    with pytest.raises(InvalidParamError):
        w.set_button_text("test", "test")
    assert w.get_button_text("back") == "test"
    with pytest.raises(InvalidParamError):
        w.get_button_text("test")
    w.get_button("back")
    with pytest.raises(InvalidParamError):
        w.get_button("test")
    w.set_title_format("auto")
    with pytest.raises(InvalidParamError):
        w.set_title_format("test")
    assert w.get_title_format() == "auto"
    w.set_subtitle_format("auto")
    with pytest.raises(InvalidParamError):
        w.set_subtitle_format("test")
    assert w.get_subtitle_format() == "auto"
    w.set_option("help_button_on_right", True)
    with pytest.raises(InvalidParamError):
        w.set_option("test", True)
    assert w.get_option("help_button_on_right") is True
    with pytest.raises(InvalidParamError):
        w.get_option("test")


def test_wizardpage(qtbot):
    page = widgets.WizardPage()
    qtbot.addWidget(page)
    pix = gui.Pixmap(100, 100)
    page.set_pixmap("background", pix)
    with pytest.raises(InvalidParamError):
        page.set_pixmap("test", pix)
    assert bytes(page.get_pixmap("background")) == bytes(pix)
    with pytest.raises(InvalidParamError):
        page.get_pixmap("test")
    page.set_button_text("back", "test")
    with pytest.raises(InvalidParamError):
        page.set_button_text("test", "test")
    assert page.get_button_text("back") == "test"
    with pytest.raises(InvalidParamError):
        page.get_button_text("test")



from __future__ import annotations

from typing import TYPE_CHECKING, Literal, Self

from prettyqt import core
from prettyqt.utils import datatypes, get_repr


if TYPE_CHECKING:
    from collections.abc import Iterator


class LineF(core.QLineF):
    """Two-dimensional vector using floating point precision."""

    def __repr__(self):
        return get_repr(self, self.get_p1(), self.get_p2())

    @property
    def _x1(self) -> float:
        return self.get_x1()

    @property
    def _y1(self) -> float:
        return self.y1()

    @property
    def _x2(self) -> float:
        return self.x2()

    @property
    def _y2(self) -> float:
        return self.y2()

    __match_args__ = ("_x1", "_y1", "_x2", "_y2")

    def __reduce__(self):
        return type(self), (self.get_p1(), self.get_p1())

    def __reversed__(self):
        return LineF(self.get_p2(), self.get_p1())

    def __abs__(self) -> float:
        return self.length()

    def __iter__(self) -> Iterator[core.PointF]:
        yield self.get_p1()
        yield self.get_p2()

    def __getitem__(self, index: Literal[0, 1]) -> core.PointF:
        match index:
            case 0:
                return self.get_p1()
            case 1:
                return self.get_p2()
            case _:
                raise KeyError(index)

    def __setitem__(self, index: Literal[0, 1], value: datatypes.PointFType):
        match index:
            case 0:
                self.set_p1(value)
            case 1:
                self.set_p2(value)
            case _:
                raise KeyError(index)

    def set_p1(self, point: datatypes.PointFType):
        self.setP1(datatypes.to_pointf(point))

    def get_p1(self) -> core.PointF:
        return core.PointF(self.p1())

    def set_p2(self, point: datatypes.PointFType):
        self.setP2(datatypes.to_pointf(point))

    def get_p2(self) -> core.PointF:
        return core.PointF(self.p2())

    def get_center(self) -> core.PointF:
        return core.PointF(self.center())

    def get_normal_vector(self) -> Self:
        return type(self)(self.normalVector())

    def get_unit_vector(self) -> Self:
        return type(self)(self.unitVector())

    def to_line(self) -> core.Line:
        return core.Line(self.toLine())


if __name__ == "__main__":
    line = LineF(core.Point(0, 0), core.Point(2, 2))



from __future__ import annotations

import io
import logging

from lxml import etree

from prettyqt import core, itemmodels


logger = logging.getLogger(__name__)


class LxmlModel(itemmodels.BaseXmlModel):
    """DOM xml model based on lxml. Parses full xml.

    Covers more features than the lazy models. (modifying the tree, xpath, ..)
    """

    SUPPORTS = bytes | str | etree.ElementTree

    def __init__(
        self,
        obj: str | bytes | etree.ElementTree,
        **kwargs,
    ):
        match obj:
            case bytes():
                xml_str = etree.XML(obj.decode())
                tree = etree.ElementTree(xml_str)
            case str():
                tree = etree.ElementTree(etree.XML(obj))
            case etree.ElementTree():
                tree = obj
            case _:
                raise TypeError(obj)
        self.tree = tree
        super().__init__(obj=self.tree.getroot(), **kwargs)

    @classmethod
    def supports(cls, instance) -> bool:
        match instance:
            case etree.ElementTree():
                return True
            case _:
                return False

    def get_parent_node(self, node_or_index: core.ModelIndex | etree.Element):
        if not isinstance(node_or_index, core.ModelIndex):
            return node_or_index.getparent()
        index = node_or_index
        parent = index.parent()
        return parent.data(self.Roles.NodeRole)

    @staticmethod
    def get_key_for_element(element: etree.Element) -> list[int]:
        child_elem = element
        key = []
        while (parent_elem := child_elem.getparent()) is not None:
            key.append(parent_elem.index(child_elem))
            child_elem = parent_elem
        return key

    def get_indexes_from_xpath(self, xpath: str):
        elements = self.tree.xpath(xpath)
        keys = [self.get_key_for_element(element) for element in elements]
        [self.index_from_key(key) for key in keys]


class LazyLxmlModel(itemmodels.BaseXmlModel):
    """Semi-lazy xml model based on lxml. Fetches all direct child nodes when needed.

    Model cant be modified, that only really makes sense for a full DOM implementation.
    """

    SUPPORTS = io.BytesIO | bytes | str | etree.iterparse | etree.Element

    def __init__(
        self,
        obj: str | bytes | io.BytesIO | etree.ElementTree,
        tag: str | None = None,
        **kwargs,
    ):
        match obj:
            case io.BytesIO():
                context = etree.iterparse(obj, events=("start",), tag=tag)
                _, root = next(context)
            case bytes():
                context = etree.iterparse(io.BytesIO(obj), events=("start",), tag=tag)
                _, root = next(context)
            case str():
                context = etree.iterparse(
                    io.BytesIO(obj.encode()), events=("start",), tag=tag
                )
                _, root = next(context)
            case etree.iterparse():
                _, root = next(obj)
            case etree.Element():
                root = obj
            case _:
                raise TypeError(obj)
        super().__init__(obj=root, **kwargs)

    @classmethod
    def supports(cls, instance) -> bool:
        match instance:
            case etree.Element() | etree.iterparse():
                return True
            case _:
                return False

    def get_parent_node(self, node_or_index: core.ModelIndex | etree.Element):
        if not isinstance(node_or_index, core.ModelIndex):
            return node_or_index.getparent()
        index = node_or_index
        parent = index.parent()
        return parent.data(self.Roles.NodeRole)

    def get_key_for_element(self, element):
        child_elem = self
        key = []
        while parent_elem := child_elem.getparent():
            key.append(parent_elem.index(child_elem))
            child_elem = parent_elem
        return key


if __name__ == "__main__":
    from prettyqt import widgets

    app = widgets.app()
    xml = """<root>
<element key='value'>text</element>
<element><sub>text</sub></element>
<empty-element xmlns="http://testns/" />
</root>
"""
    table = widgets.TreeView()
    model = LxmlModel(xml, parent=table)
    table.setRootIsDecorated(True)
    # table.setSortingEnabled(True)
    table.set_model(model)

    def test(new, old):
        node = new.data(new.model().Roles.NodeRole)
        new.model().get_parent_node(node)

    table.selectionModel().currentChanged.connect(test)
    print(model.get_indexes_from_xpath("element"))
    table.show()
    with app.debug_mode():
        app.exec()



from __future__ import annotations

from typing import Literal

from prettyqt import bluetooth, core
from prettyqt.utils import bidict


module = bluetooth.QBluetoothServiceDiscoveryAgent

DiscoveryModeStr = Literal["minimal", "full"]

DISCOVERY_MODES: bidict[DiscoveryModeStr, module.DiscoveryMode] = bidict(
    minimal=module.DiscoveryMode.MinimalDiscovery,
    full=module.DiscoveryMode.FullDiscovery,
)

ErrorStr = Literal[
    "none", "powered_off", "input_output", "invalid_bluetooth_adapter", "unknown"
]

ERRORS: bidict[ErrorStr, module.Error] = bidict(
    none=module.Error.NoError,
    powered_off=module.Error.PoweredOffError,
    input_output=module.Error.InputOutputError,
    invalid_bluetooth_adapter=module.Error.InvalidBluetoothAdapterError,
    unknown=module.Error.UnknownError,
)


class BluetoothServiceDiscoveryAgent(
    core.ObjectMixin, bluetooth.QBluetoothServiceDiscoveryAgent
):
    """Enables you to query for Bluetooth services."""

    def start_discovery(self, full: bool = False):
        """Start bluetooth service discovery.

        Arguments:
            full: full discovery instead of minimal discovery.
        """
        if full:
            flag = bluetooth.QBluetoothServiceDiscoveryAgent.FullDiscovery
        else:
            flag = bluetooth.QBluetoothServiceDiscoveryAgent.MinimalDiscovery
        self.start(flag)

    def get_error(self) -> ErrorStr:
        """Get error code."""
        return ERRORS.inverse[self.error()]

    def set_remote_address(
        self, address: str | int | bluetooth.QBluetoothAddress
    ) -> bool:
        """Set remote address.

        Arguments:
            address: address of the remote
        """
        address = bluetooth.BluetoothAddress(address)
        return self.setRemoteAddress(address)

    def get_discovered_services(self) -> list[bluetooth.BluetoothServiceInfo]:
        """Get list of discovered devices."""
        return [bluetooth.BluetoothServiceInfo(i) for i in self.discoveredServices()]



from __future__ import annotations

from typing import TYPE_CHECKING

from prettyqt import constants, core, eventfilters


if TYPE_CHECKING:
    from prettyqt import widgets


class SectionAutoSpanEventFilter(eventfilters.BaseEventFilter):
    ID = "autospan_sections"

    def __init__(
        self,
        parent: widgets.TableView | widgets.TreeView,
        orientation: (
            constants.Orientation | constants.OrientationStr
        ) = constants.HORIZONTAL,
    ):
        super().__init__(parent)
        self._widget = parent
        self._last_start = None
        self._last_end = None
        self.orientation = constants.ORIENTATION.get_enum_value(orientation)
        if self.orientation == constants.HORIZONTAL:
            parent.h_scrollbar.valueChanged.connect(self._update_spans)
        else:
            parent.v_scrollbar.valueChanged.connect(self._update_spans)
        parent.model().headerDataChanged.connect(self._on_force_update)
        parent.model_changed.connect(self._on_force_update)
        parent.installEventFilter(self)
        self._update_spans()
        # sel_model = parent.selectionModel()
        # if not sel_model:
        #     return

    def _on_force_update(self):
        self._update_spans(True)

    def eventFilter(self, obj, event: core.Event) -> bool:
        match event.type():
            case core.Event.Type.Resize:
                self._update_spans()
                return False
        return super().eventFilter(obj, event)

    def _update_spans(self, force: bool = False):
        cols = self._widget.get_visible_section_span("horizontal", margin=1)
        rows = self._widget.get_visible_section_span("vertical", margin=1)
        start = (rows[0], cols[0])
        end = (rows[1], cols[1])
        if start == self._last_start and end == self._last_end and not force:
            return
        self._last_start = start
        self._last_end = end
        self._widget.clearSpans()
        self._widget.auto_span(orientation=self.orientation, start=start, end=end)


if __name__ == "__main__":
    from prettyqt import debugging, widgets

    app = widgets.app()
    widget = debugging.example_table()
    widget.set_delegate("editor")
    with app.debug_mode():
        test = SectionAutoSpanEventFilter(widget)
        widget.show()
        app.exec()



"""Tests for `prettyqt` package."""

import pytest

from prettyqt import constants
from prettyqt.qt import QtCore
from prettyqt.utils import InvalidParamError


charts = pytest.importorskip("prettyqt.charts")


def test_barseries(qtbot):
    series = charts.BarSeries()
    barset = charts.BarSet("Test")
    series.append(barset)
    assert series[0] == barset
    del series[0]
    series.set_labels_position("inside_base")
    with pytest.raises(InvalidParamError):
        series.set_labels_position("test")
    assert series.get_labels_position() == "inside_base"


def test_barcategoryaxis(qtbot):
    obj = QtCore.QObject()
    axis = charts.BarCategoryAxis(obj)
    axis += "a"
    axis += "b"
    assert axis[0] == "a"
    axis["a"] = "c"
    assert list(axis) == ["c", "b"]
    del axis["b"]


def test_boxset(qtbot):
    boxset = charts.BoxSet()
    repr(boxset)
    boxset.get_pen()
    boxset.get_brush()
    boxset["median"] = 1.0
    assert boxset["median"] == 1.0


def test_candlestickset(qtbot):
    candlestickset = charts.CandlestickSet()
    repr(candlestickset)
    candlestickset.get_pen()
    candlestickset.get_brush()


def test_categoryaxis(qtbot):
    axis = charts.CategoryAxis()
    axis += ("a", 2)
    axis += ("b", 4)
    assert len(axis) == 2
    assert axis[0] == "a"
    axis["a"] = "c"
    assert list(axis) == ["c", "b"]
    del axis["c"]
    axis.set_labels_position("on_value")
    with pytest.raises(InvalidParamError):
        axis.set_labels_position("test")
    assert axis.get_labels_position() == "on_value"


def test_chart(qtbot):
    chart = charts.Chart()
    chart.hide_legend()
    chart.show_legend()
    chart.set_legend_alignment("right")
    chart.set_theme("Dark")
    chart.set_animation_options("series")


def test_chartview(qtbot, qttester):
    widget = charts.ChartView()
    widget.show()
    widget.get_image()
    qtbot.add_widget(widget)
    qttester.send_keypress(widget, constants.Key.Key_F11)
    # qttester.send_keypress(widget, constants.Key.Key_Minus)
    # qttester.send_keypress(widget, constants.Key.Key_Plus)
    qttester.send_keypress(widget, constants.Key.Key_Left)
    qttester.send_keypress(widget, constants.Key.Key_Right)
    qttester.send_keypress(widget, constants.Key.Key_Up)
    qttester.send_keypress(widget, constants.Key.Key_Down)
    qttester.send_mousepress(widget, constants.MouseButton.RightButton)
    qttester.send_mousemove(widget, delay=100)


def test_legend(qtbot):
    chart = charts.Chart()
    legend = chart.get_legend()
    legend.set_alignment("bottom")
    with pytest.raises(InvalidParamError):
        legend.set_alignment("test")
    assert legend.get_alignment() == "bottom"
    legend.set_marker_shape("circle")
    with pytest.raises(InvalidParamError):
        legend.set_marker_shape("test")
    assert legend.get_marker_shape() == "circle"
    legend.get_border_color()
    legend.get_color()
    legend.get_label_color()
    legend.get_font()


def test_lineseries(qtbot):
    line = charts.LineSeries()
    line += QtCore.QPointF(1, 1)
    line += QtCore.QPointF(2, 2)
    line[1] = QtCore.QPointF(0, 0)
    line.get_brush()
    line.get_pen()
    del line[1]


def test_datetimeaxis(qtbot):
    axis = charts.DateTimeAxis()
    axis.get_min()
    axis.get_max()


def test_scatterseries(qtbot):
    charts.ScatterSeries()


def test_pieslice(qtbot):
    pieslice = charts.PieSlice()
    repr(pieslice)
    pieslice.set_label_position("inside_tangential")
    with pytest.raises(InvalidParamError):
        pieslice.set_label_position("test")
    assert pieslice.get_label_position() == "inside_tangential"
    pieslice.get_label_font()
    pieslice.get_label_brush()
    pieslice.get_label_color()
    pieslice.get_pen()
    pieslice.get_brush()
    pieslice.get_color()
    pieslice.get_border_color()


def test_polarchart(qtbot):
    chart = charts.PolarChart()
    chart.add_axis(charts.ValueAxis(), "radial")


def test_valueaxis(qtbot):
    axis = charts.ValueAxis()
    axis.set_tick_type("fixed")
    with pytest.raises(InvalidParamError):
        axis.set_tick_type("test")
    assert axis.get_tick_type() == "fixed"
    assert axis.get_alignment() is None
    assert axis.get_orientation() is None
    # assert axis.get_orientation() == "horizontal"
    axis.get_grid_line_color()
    axis.get_grid_line_pen()
    axis.get_line_pen()
    axis.get_line_pen_color()
    axis.get_labels_color()
    axis.get_labels_brush()
    axis.get_labels_font()
    axis.get_title_font()
    axis.get_title_brush()
    axis.get_shades_color()
    axis.get_shades_brush()
    axis.get_shades_pen()
    axis.get_shades_border_color()
    axis.get_minor_grid_line_pen()
    axis.get_minor_grid_line_color()
